'''
Check call-throughs.
'''

import numbers
import numpy as np
import pytest
import scipy

import mqr

def test_test():
    x = np.array([1, 2, 3, 5, 5])
    y = np.array([1, 2, 2, 3, 5])
    alternative = 'two-sided'

    method = 'spearman'
    res = mqr.inference.nonparametric.correlation.test(x, y, alternative, method)
    assert res.description == 'correlation coefficient'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'corr(x, y)'
    assert res.sample_stat_target == 0.0
    assert isinstance(res.sample_stat_value, numbers.Number)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    method = 'kendall'
    res = mqr.inference.nonparametric.correlation.test(x, y, alternative, method)
    assert res.description == 'correlation coefficient'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'corr(x, y)'
    assert res.sample_stat_target == 0.0
    assert isinstance(res.sample_stat_value, numbers.Number)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)



import IPython.core.formatters

import mqr
from mqr.notebook.defaults import Defaults
from mqr.notebook import formatters

try:
    IPython.get_ipython().display_formatter.formatters["text/html"] = formatters.HTMLFormatter()
except:
    pass



from mqr.inference.confint import ConfidenceInterval
from mqr.inference.hyptest import HypothesisTest
from mqr.inference.power import TestPower

import mqr.inference.lib.util as util
import mqr.interop.inference as interop

import numpy as np
import scipy
import statsmodels

def power_1sample(effect, nobs, alpha=0.05, alternative='two-sided', method='chi2'):
    '''
    Calculate power for test of variance of a sample.

    Null-hypothesis
        `effect` = var / var_0 == 1

    Parameters
    ----------
    effect : float
        Effect size.
    nobs : float
        Number of observations in sample.
    alpha : float, optional
        Significance level.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'chi2'}, optional
        Only the 'chi2' is available. Calculates the power directly from the
        Chi-squared distribution.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    '''
    den = effect
    dist = scipy.stats.chi2(nobs-1)

    if alternative == 'less':
        num = dist.ppf(alpha)
        power = dist.cdf(num/den)
    elif alternative == 'greater':
        num = dist.ppf(1-alpha)
        power = 1 - dist.cdf(num/den)
    elif alternative == 'two-sided':
        num_1 = dist.ppf(1-alpha/2)
        num_2 = dist.ppf(alpha/2)
        power = 1 - (dist.cdf(num_1/den) - dist.cdf(num_2/den))
    else:
        raise ValueError(util.alternative_error_msg(alternative))

    return TestPower(
        name='variance',
        alpha=alpha,
        beta=1-power,
        effect=effect,
        alternative=alternative,
        method='chi2',
        sample_size=nobs)

def power_2sample(var_ratio, nobs, alpha=0.05, alternative='two-sided', method='f'):
    '''
    Calculate power for test of ratio of variances.

    Null-hypothesis
        `effect` = var_1 / var_2 == 1

    Parameters
    ----------
    var_ratio : float
        Effect size.
    nobs : float
        Number of observations in sample.
    alpha : float, optional
        Significance level.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'f'}, optional
        Only 'f' is available. Calculates the power directly from the
        F distribution.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    '''
    dist = scipy.stats.f(nobs-1, nobs-1)

    if alternative == 'less':
        num = dist.ppf(alpha)
        power = dist.cdf(num/var_ratio)
    elif alternative == 'greater':
        num = dist.ppf(1-alpha)
        power = 1 - dist.cdf(num/var_ratio)
    elif alternative == 'two-sided':
        num_1 = dist.ppf(1-alpha/2)
        num_2 = dist.ppf(alpha/2)
        power = 1 - (dist.cdf(num_1/var_ratio) - dist.cdf(num_2/var_ratio))
    else:
        raise ValueError(util.alternative_error_msg(alternative))

    return TestPower(
        name='ratio of variances',
        alpha=alpha,
        beta=1-power,
        effect=var_ratio,
        alternative=alternative,
        method='f',
        sample_size=nobs)

def size_1sample(effect, alpha, beta, alternative='two-sided', method='chi2'):
    '''
    Calculate sample size for test of variance of a sample.

    Null-hypothesis
        `effect` == var / var_0 == 1

    Parameters
    ----------
    effect : float
        Required effect size.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'chi2'}, optional
        Numerically solves :func:`power_1sample` (passing `method`) equal to
        1 - 'beta' by changing `n`. Only 'chi2' is available.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    '''
    if (effect < 1) and (alternative == 'greater'):
        raise ValueError('alternative "greater" not valid when effect < 1')
    elif (effect > 1) and (alternative == 'less'):
        raise ValueError('alternative "less" not valid when effect > 1')

    def beta_fn(n):
        power = power_1sample(
            effect=effect,
            nobs=n,
            alpha=alpha,
            alternative=alternative,
            method=method)
        return power.beta - beta
    nobs = scipy.optimize.fsolve(beta_fn, 2)[0]

    return TestPower(
        name='variance',
        alpha=alpha,
        beta=beta,
        effect=effect,
        alternative=alternative,
        method='chi2',
        sample_size=nobs)

def size_2sample(var_ratio, alpha, beta, alternative='two-sided', method='f'):
    '''
    Calculate sample size for test of ratio of variances.

    Null-hypothesis
        `var_ratio` = var_1 / var_2 == 1

    Parameters
    ----------
    var_ratio : float
        Required effect size.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'f'}, optional
        Numerically solves :func:`power_2sample` (passing `method`) equal to
        1 - 'beta' by changing `n`. Only 'f' is available.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    '''
    if (alternative == 'less') and (var_ratio > 1.0):
            raise ValueError('diff must be > 1 for a greater-than alternative hypothesis')
    elif (alternative == 'greater') and (var_ratio < 1.0):
            raise ValueError('diff must be < 1 for a less-than alternative hypothesis')

    def beta_fn(n):
        power = power_2sample(
            var_ratio=var_ratio,
            nobs=n,
            alpha=alpha,
            alternative=alternative)
        return power.beta - beta
    nobs = scipy.optimize.fsolve(beta_fn, 2)[0]

    return TestPower(
        name='ratio of variances',
        alpha=alpha,
        beta=beta,
        effect=var_ratio,
        alternative=alternative,
        method='f',
        sample_size=nobs)

def confint_1sample(x, conf=0.95, bounded='both', method='chi2'):
    '''
    Confidence interval for the variance of a sample.

    Parameters
    ----------
    x : array_like
        Calcaulate interval for the variance of this sample.
    conf : float, optional
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}, optional
        Which sides of the interval to close.
    method : {'chi2'}, optional
        Only 'chi2' for an exact interval based on the chi-squared distribution.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`
    '''
    if method == 'chi2':
        alpha = 1 - conf
        nobs = len(x)
        dof = nobs - 1
        s2 = np.var(x, ddof=1)
        dist = scipy.stats.chi2(dof)
        if bounded == 'both':
            lower, upper = dof * s2 / dist.ppf([1 - alpha / 2, alpha / 2])
        elif bounded == 'below':
            lower = dof * s2 / dist.ppf(1 - alpha)
            upper = np.inf
        elif bounded == 'above':
            lower = 0.0
            upper = dof * s2 / dist.ppf(alpha)
        else:
            raise ValueError(util.bounded_error_msg(bounded))
    else:
        raise ValueError(util.method_error_msg(method, ['chi2']))

    return ConfidenceInterval(
        name='variance',
        method=method,
        value=s2,
        lower=lower,
        upper=upper,
        conf=conf,
        bounded=bounded)

def confint_2sample(x, y, conf=0.95, bounded='both', method='f'):
    '''
    Confidence interval for the ratio of variances of two samples.

    Parameters
    ----------
    x, y : array_like
        Calcaulate interval for the ratio of variances of these two samples.
    conf : float, optional
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}, optional
        Which sides of the interval to close.
    method : {'f'}, optional
        Only 'f' for an exact interval based on the F distribution.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`
    '''
    if method == 'f':
        alpha = 1 - conf
        s2x = np.var(x, ddof=1)
        s2y = np.var(y, ddof=1)
        nobsx = len(x)
        nobsy = len(y)
        dofx = nobsx - 1
        dofy = nobsy - 1
        ratio = s2x / s2y
        dist = scipy.stats.f(dofy, dofx)
        if bounded == 'both':
            lower, upper = ratio * dist.ppf([alpha / 2, 1 - alpha / 2])
        elif bounded == 'below':
            lower = ratio * dist.ppf(alpha)
            upper = np.inf
        elif bounded == 'above':
            lower = -np.inf
            upper = ratio * dist.ppf(1 - alpha)
        else:
            raise ValueError(util.bounded_error_msg(bounded))
    else:
        raise ValueError(util.method_error_msg(method, ['f']))

    return ConfidenceInterval(
        name='ratio of variances',
        method=method,
        value=ratio,
        lower=lower,
        upper=upper,
        conf=conf,
        bounded=bounded)

def test_1sample(x, H0_var, alternative='two-sided', method='chi2'):
    '''
    Hypothesis test for the varianve of a sample.

    Null hypothesis
        var(`x`) / `H0_var` == 1

    Parameters
    ----------
    x : array_like
        Test variance of this sample.
    H0_var : float
        Null-hypothesis variance.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'chi2'}, optional
        Only 'chi2' for an exact test based on the Chi-squared distribution.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    '''
    if method == 'chi2':
        dfx = len(x) - 1
        s2x = np.var(x, ddof=1)
        q = dfx * s2x / H0_var # Eqn 8-17, MR
        dist = scipy.stats.chi2(dfx)

        stat = s2x
        if alternative == 'less':
            pvalue = dist.cdf(q)
        elif alternative == 'greater':
            pvalue = 1 - dist.cdf(q)
        elif alternative == 'two-sided':
            pvalue = 2.0 * np.minimum(dist.cdf(q), 1-dist.cdf(q))
        else:
            raise ValueError(util.alternative_error_msg(alternative))
    else:
        raise ValueError(util.method_error_msg(method, ['chi2']))

    x_name = util.var_name(x, 'x')
    return HypothesisTest(
        description='variance',
        alternative=alternative,
        method='chi2',
        sample_stat=f'var({x_name})',
        sample_stat_target=H0_var,
        sample_stat_value=s2x,
        stat=stat,
        pvalue=pvalue,)

def test_2sample(x, y, alternative='two-sided', method='f'):
    '''
    Hypothesis test for the ratio of variances of two samples.

    Null hypothesis
        var(`x`) / var(`y`) == 1

    Parameters
    ----------
    x, y : array_like
        Test ratio of variances of these two samples.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'f', 'bartlett'}, optional
        | 'f'
        |   F-test from the F distribution of the ratio of equal variances.
        | 'bartlett'
        |   Bartlett's test.
            Calls :func:`scipy..bartlett <scipy.stats.bartlett>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    '''
    if method == 'f':
        description = 'ratio of variances'
        tgt = 1.0
        rel = '/'

        dfx = len(x) - 1
        dfy = len(y) - 1
        f = np.var(x, ddof=1) / np.var(y, ddof=1)
        dist = scipy.stats.f(dfx, dfy)

        stat = f
        if alternative == 'less':
            pvalue = dist.cdf(f)
        elif alternative == 'greater':
            pvalue = 1 - dist.cdf(f)
        elif alternative == 'two-sided':
            pvalue = 2.0 * np.minimum(1.0 - dist.cdf(f), dist.cdf(f))
        else:
            raise ValueError(util.alternative_error_msg(alternative))
    elif method == 'levene':
        if alternative != 'two-sided':
            raise ValueError('one-sided alternative not available in Levene test')
        description = 'equality of variances'
        tgt = 0.0
        rel = '-'
        (stat, pvalue) = scipy.stats.levene(x, y)
    elif method == 'bartlett':
        if alternative != 'two-sided':
            raise ValueError('one-sided alternative not available in Bartlett test')
        description = 'equality of variances'
        tgt = 0.0
        rel = '-'
        (stat, pvalue) = scipy.stats.bartlett(x, y)
    else:
        raise ValueError(util.method_error_msg(method, ['f', 'bartlett']))

    x_name = util.var_name(x, 'x')
    y_name = util.var_name(y, 'y')
    return HypothesisTest(
        description=description,
        alternative=alternative,
        method=method,
        sample_stat=f'var({x_name}) {rel} var({y_name})',
        sample_stat_target=tgt,
        sample_stat_value=stat,
        stat=stat,
        pvalue=pvalue,)



'''
Check call-throughs.
'''

import numbers
import numpy as np
import pytest
import scipy

import mqr

def test_test_nsample():
    x, y, z = np.array([
       [0.36522057, 0.37377119, 0.86150726, 0.96718967, 0.47966424,
        0.74979588, 0.35591634, 0.75272332, 0.17909715, 0.66216129],
       [0.30579918, 0.01280289, 0.2218879 , 0.77332064, 0.18087063,
        0.85439362, 0.41027609, 0.96153719, 0.94262441, 0.88509784],
       [0.1558813 , 0.25968253, 0.32619182, 0.89433078, 0.51197809,
        0.5814941 , 0.60785036, 0.32729355, 0.68637953, 0.9616885 ]], dtype=float)
    alternative = 'two-sided'

    method = 'levene'
    res = mqr.inference.nonparametric.variance.test_nsample(x, y, z, alternative=alternative, method=method)
    assert res.description == 'equality of variances'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'var(x_i)'
    assert res.sample_stat_target == 'var(x_j)'
    assert isinstance(res.sample_stat_value, numbers.Number)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    method = 'fligner-killeen'
    res = mqr.inference.nonparametric.variance.test_nsample(x, y, z, alternative=alternative, method=method)
    assert res.description == 'equality of variances'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_stat == 'var(x_i)'
    assert res.sample_stat_target == 'var(x_j)'
    assert isinstance(res.sample_stat_value, numbers.Number)
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)



import mqr

import numpy as np
import pandas as pd
import pytest

@pytest.fixture
def grr_data():
    design = mqr.doe.Design.from_fullfact(
        names=['Part', 'Inspector'],
        levels=[10, 3],
        scale_origin=False)
    design = design.replicate(3, label='Test')
    design.levels = (design.levels + 1).astype(int)
    data = design.to_df()
    data['Impedance'] = np.array([
        37, 38, 37, 42, 41, 43, 30, 31, 31, 42, 43, 42, 28, 30, 29, 42, 42, 43, 25, 26, 27, 40, 40, 40, 25, 25, 25, 35, 34, 34,
        41, 41, 40, 42, 42, 42, 31, 31, 31, 43, 43, 43, 29, 30, 29, 45, 45, 45, 28, 28, 30, 43, 42, 42, 27, 29, 28, 35, 35, 34,
        41, 42, 41, 43, 42, 43, 29, 30, 28, 42, 42, 42, 31, 29, 29, 44, 46, 45, 29, 27, 27, 43, 43, 41, 26, 26, 26, 35, 34, 35,
    ])
    return data

def test_NameMapping_init():
    name_mapping = mqr.msa.NameMapping()
    name_mapping.measurement == 'measurement'
    name_mapping.part == 'part'
    name_mapping.operator == 'operator'

    name_mapping = mqr.msa.NameMapping(
        measurement='m',
        part='p',
        operator='o',)
    name_mapping.measurement == 'm'
    name_mapping.part == 'p'
    name_mapping.operator == 'o'

def test_GRR_init(grr_data):
    names = mqr.msa.NameMapping(
        part='Part',
        operator='Inspector',
        measurement='Impedance',)
    grr = mqr.msa.GRR(grr_data, 3.0, names, include_interaction=True)
    assert grr.formula == 'Impedance ~ C(Part) + C(Inspector) + C(Part):C(Inspector)'
    assert list(grr.counts) == [10, 3]
    assert grr.replicates == 3
    assert grr.model
    assert grr.regression_result

    grr = mqr.msa.GRR(grr_data, 3.0, names, include_interaction=False)
    assert grr.formula == 'Impedance ~ C(Part) + C(Inspector)'
    assert list(grr.counts) == [10, 3]
    assert grr.replicates == 3
    assert grr.model
    assert grr.regression_result

def test_VarianceTable_init(grr_data):
    names = mqr.msa.NameMapping(
        part='Part',
        operator='Inspector',
        measurement='Impedance',)
    grr = mqr.msa.GRR(grr_data, 3.0, names)
    var = mqr.msa.VarianceTable(grr)
    varcomp = var.table['VarComp']

    assert var.grr == grr

def test_VarianceTable_varcomp(grr_data):
    """
    Verifies variance components against Table 8.8 in [1]_.

    References
    ----------
    .. [1]  Montgomery, D. C. (2009).
            Statistical quality control (Vol. 7).
            New York: Wiley.
    """
    names = mqr.msa.NameMapping(
        part='Part',
        operator='Inspector',
        measurement='Impedance',)
    grr = mqr.msa.GRR(grr_data, 3.0, names)
    var = mqr.msa.VarianceTable(grr)
    varcomp = var.table['VarComp']
    
    assert varcomp['Part-to-Part'] == pytest.approx(48.2926, abs=1e-4)
    assert varcomp['Operator'] == pytest.approx(0.5646, abs=1e-4)
    assert varcomp['Operator*Part'] == pytest.approx(0.7280, abs=1e-4)
    assert varcomp['Repeatability'] == pytest.approx(0.5111, abs=1e-4)
    assert varcomp['Gauge RR'] + varcomp['Part-to-Part'] == varcomp['Total']
    assert varcomp['Repeatability'] + varcomp['Reproducibility'] == varcomp['Gauge RR']
    assert varcomp['Operator'] + varcomp['Operator*Part'] == varcomp['Reproducibility']

    grr = mqr.msa.GRR(grr_data, 3.0, names, include_interaction=False)
    var = mqr.msa.VarianceTable(grr)
    varcomp = var.table['VarComp']

    assert varcomp['Gauge RR'] + varcomp['Part-to-Part'] == varcomp['Total']
    assert varcomp['Repeatability'] + varcomp['Reproducibility'] == varcomp['Gauge RR']
    assert varcomp['Operator'] == varcomp['Reproducibility']

def test_VarianceTable_set_discrimination(grr_data):
    names = mqr.msa.NameMapping(
        part='Part',
        operator='Inspector',
        measurement='Impedance',)
    grr = mqr.msa.GRR(grr_data, 3.0, names)
    var = mqr.msa.VarianceTable(grr)

    assert var.discrimination == pytest.approx(54.5, abs=1e-1)

def test_VarianceTable_set_num_distinct_cats(grr_data):
    names = mqr.msa.NameMapping(
        part='Part',
        operator='Inspector',
        measurement='Impedance',)
    grr = mqr.msa.GRR(grr_data, 3.0, names)
    var = mqr.msa.VarianceTable(grr)

    assert round(var.num_distinct_cats) == 7



"""
================================================
Correlation matrix (:mod:`mqr.plot.correlation`)
================================================

.. currentmodule:: mqr.plot.correlation

.. rubric:: Functions
.. autosummary::
    :toctree: generated/

    matrix
"""
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns

from mqr import inference
from mqr.plot.defaults import Defaults
from mqr.plot.lib.util import set_kws

def matrix(data, ax, conf=0.95, fig=None, show_conf=False, cmap='RdBu',
           hist_kws=None, reg_kws=None, scatter_kws=None, line_kws=None, text_kws=None):
    """
    Plot a correlation matrix and associated statistics.

    Plots an n by n grid (where n is the number of columns in data). Each column
    and row corresponds to a column in the dataframe `data`. The lower diagonal
    shows scatter plots between the corresponding variables. The upper diagonal
    shows the statistics (1) Pearson correlation coefficient, (2) p-value for
    the correlation coefficient, and when `show_conf` is true, (3) confidence
    interval for the correlation coefficient.

    Also:

    - Whenever the p-value is less than the significance level (`1-conf`), the
        values are printed in bold.
    - When `show_conf` is true, the upper triangle is also coloured according to
        the correlation coefficient. Positive correlations are shown in shades
        of blue while negative correlations are shown in shades of red. The
        saturation of the shade corresponds to the significance level of the
        correlation.

    Parameters
    ----------
    data : pandas.DataFrame
        Data with samples in columns. Calculate the correlations between columns.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    conf : float, optional
        Confidence level of the interval.
    fig : matplotlib.figure.Figure, optional
        The figure to which the axes belong. If not provided, the figure will be
        taken from the first axes.
    show_conf : bool, optional
        Shows confidence intervals and colours when true.
    cmap : str, optional
        Name of a colormap. Correlation coefficient is mapped to a colour in
        this colormap and shown in the upper triangle.
    hist_kws : dict, optional
        Keyword args for the histogram. Passed to :func:`seaborn.histplot`.
    reg_kws : dict, optional
        Keyword args for the regression result.
        Passed as ``**reg_kws`` to :func:`seaborn.regplot`.
    scatter_kws : dict, optional
        Keyword args for the scatter plot.
        Passed as ``scatter_kws=scatter_kws`` to :func:`seaborn.regplot`.
    line_kws : dict, optional
        Keyword args for the line of best fit.
        Passed as ``line_kws=line_kws`` to :func:`seaborn.regplot`.
    text_kws : dict, optional
        Keyword args for the text in the upper-right entries.
        Passed as ``**text_kws`` to :func:`matplotlib.pyplot.text`.

    Examples
    --------
    Load sample data and show correlations, including confidence intervals.

    .. plot::

        data = pd.read_csv(mqr.sample_data('study-random-5x5.csv'))

        fig, ax = plt.subplots(5, 5, figsize=(6, 6))
        mqr.plot.correlation.matrix(
            data.loc[:, 'KPI1':'KPO2'],
            show_conf=True,
            ax=ax,
            fig=fig)
        plt.show()    

    """
    M, N = data.shape

    assert ax.shape == (N, N) , f'Tableau shape is ({N},{N}) but ax.shape is {ax.shape}.'
    assert data.ndim == 2, '`data` must be a 2-dimensional array.'

    colormap = mpl.colormaps[cmap]

    hist_kws = set_kws(
        hist_kws,
        color='C0',
        stat='count',
    )
    reg_kws = set_kws(
        reg_kws,
        marker='.',
    )
    scatter_kws = set_kws(
        scatter_kws,
        color='C0',
    )
    line_kws = set_kws(
        line_kws,
        color='C0',
    )
    text_kws = set_kws(
        text_kws,
        color='k',
        fontsize=8,
        ha='center',
        va='center',
    )

    if fig is None:
        fig = ax[0, 0].get_figure()

    fig.set_layout_engine(None)
    fig.subplots_adjust(wspace=0, hspace=0)

    alpha = 1 - conf
    axis_names = data.columns #[name for name in study.samples]

    # Plot histograms
    for n in range(N):
        ax_hist = ax[n, n].twinx()
        sns.histplot(data.iloc[:, n], ax=ax_hist, **hist_kws)
        ax_hist.set_ylabel(None)
        ax_hist.set_yticks([])


    for i in range(N):
        for j in range(N):
            ax[i, j].set_xlabel(None)
            ax[i, j].set_ylabel(None)
            if i >= j: continue

            if show_conf:
                ci = inference.correlation.confint(
                    x=data.iloc[:, i],
                    y=data.iloc[:, j],
                    conf=conf)
            test = inference.correlation.test(
                x=data.iloc[:, i],
                y=data.iloc[:, j])
            rho = test.sample_stat_value
            p = test.pvalue

            text = f'r={rho:.2f}\n(p={p:.2f})'
            if show_conf:
                text += f'\n\n{ci.conf*100:.0f}% CI:\n[{ci.lower:.2f}, {ci.upper:.2f}]'
                ci = 100 * conf
                plot_alpha = 0.6 if p < alpha else 0.2
                fontweight = 'bold' if p < alpha else 'normal'
            else:
                ci = None
                plot_alpha = 0.6
                fontweight = 'normal'

            scatter_kws |= dict(alpha=plot_alpha, linewidths=0)
            line_kws |= dict(alpha=plot_alpha)
            text_kws |= dict(fontweight=fontweight)

            if show_conf:
                (r, g, b, a) = colormap((1 + rho) / 2)
                ax[i, j].set_facecolor((r, g, b, a))
            ax[i, j].text(
                0.5, 0.5,
                text,
                transform=ax[i, j].transAxes,
                **text_kws)

            sns.regplot(
                x=data.iloc[:, i],
                y=data.iloc[:, j],
                x_ci='ci', ci=ci,
                ax=ax[j, i],
                scatter_kws=scatter_kws,
                line_kws=line_kws,
                **reg_kws)

    # Share axes along rows and down columns ...
    for i in range(1, N):
        for j in range(1, N):
            ax[i, j].sharey(ax[i, 0]) # share y along rows

    for j in range(0, N-1):
        for i in range(1, N):
            ax[N-i-1, j].sharex(ax[N-1, j]) # share x axis up columns

    # ... turn off ticks and labels to start.
    for i in range(N-1):
        for j in range(N):
            plt.setp(ax[i, j].get_xticklabels(), visible=False)
            ax[i, j].tick_params(axis='x', bottom=False)
    for i in range(N):
        for j in range(1, N):
            plt.setp(ax[i, j].get_yticklabels(), visible=False)
            ax[i, j].tick_params(axis='y', left=False)
    plt.setp(ax[0, 0].get_yticklabels(), visible=False)
    ax[0, 0].tick_params(axis='y', left=False)

    # Rotate the tick labels along the bottom.
    for j in range(N):
        # ... along the bottom ...
        ax[N-1, j].tick_params(axis='x', rotation=90)

    # Set labels.
    for i in range(N):
        ax[i, N-1].set_ylabel(axis_names[i])
        ax[i, N-1].yaxis.set_label_position('right')

    for j in range(N):
        ax[0, j].set_xlabel(axis_names[j])
        ax[0, j].xaxis.set_label_position('top')



"""
=================================================
Measurement system analysis (:mod:`mqr.plot.msa`)
=================================================

.. currentmodule:: mqr.plot.msa



.. rubric:: Functions
.. autosummary::
    :toctree: generated/

    grr
    bar_var_pct
    box_measurement_by_part
    xbar_operator
    box_measurement_by_operator
    r_operator
    line_part_operator_intn
"""

import numpy as np
import seaborn as sns

import mqr
from mqr.plot.defaults import Defaults
from mqr.plot.lib.util import set_kws

def bar_var_pct(grr_table, ax, sources=None, bar_kws=None):
    """
    Bar graph of percent contributions from `sources` in a GRR study.

    Parameters
    ----------
    grr_table : :class:`mqr.msa.VarianceTable`
        Results from a GRR study.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    sources : list[str], optional
        Sources to include. Default `["% Contribution", "% StudyVar",
        "% Tolerance"]`.
    """
    bar_kws = set_kws(
        bar_kws,
        zorder=1,
    )

    if sources is None:
        indices = ['Gauge RR', 'Repeatability', 'Reproducibility', 'Part-to-Part']
    else:
        indices = sources
    columns = ['% Contribution', '% StudyVar', '% Tolerance']

    x = np.arange(len(indices))  # the label locations
    width = 0.8 / len(columns)  # the width of the bars
    pct_data = grr_table.table.loc[indices, columns]
    for i, row in enumerate(pct_data.items()):
        offset = width * (i - 1)
        rects = ax.bar(x + offset, row[1], width, label=row[0], **bar_kws)

    ax.legend(
        [c for c in columns],
        prop={'size': 8},
        fancybox=True,
        bbox_to_anchor=(1.02, 0.5, 0.0, 0.0),
        loc='center left',
        borderaxespad=0.0)
    ax.set_xticks(x)
    ax.set_xticklabels(indices)
    for label in ax.get_xticklabels():
        label.set_rotation(15)
        label.set_ha('right')
    ax.set_ylabel('Percentage (%)')
    ax.set_title('Components of Variation')
    ax.grid()

def box_measurement_by_part(grr, ax, box_kws=None, line_kws=None):
    """
    Box plot showing spread in a GRR by part.

    Parameters
    ----------
    grr : mqr.msa.GRR
        Results from a GRR study.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    """
    box_kws = set_kws(
        box_kws,
        color='C0',
        fill=False,
        width=0.4,
    )
    line_kws = set_kws(
        line_kws,
        color='C0',
        marker=Defaults.marker,
    )

    name_p = grr.names.part
    name_m = grr.names.measurement

    sns.boxplot(grr.data, x=name_p, y=name_m, ax=ax, **box_kws)

    names = grr.data[name_p].unique().astype('str')
    means = grr.data.groupby(name_p)[name_m].mean()
    ax.plot(names, means, **line_kws)

    ax.set_xlabel(name_p)
    ax.set_ylabel(name_m)
    ax.set_title(f'{name_m} by {name_p}')

    ax.grid()

def box_measurement_by_operator(grr, ax, box_kws=None, line_kws=None):
    """
    Box plot showing spread in a GRR by operator.

    Parameters
    ----------
    grr : mqr.msa.GRR
        Results from a GRR study.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    """
    box_kws = set_kws(
        box_kws,
        color='C0',
        fill=False,
        width=0.4,
    )
    line_kws = set_kws(
        line_kws,
        color='C0',
        marker=Defaults.marker,
    )

    name_o = grr.names.operator
    name_m = grr.names.measurement

    sns.boxplot(grr.data, x=name_o, y=name_m, ax=ax, **box_kws)

    names = grr.data[name_o].unique().astype('str')
    means = grr.data.groupby(name_o)[name_m].mean()
    ax.plot(names, means, **line_kws)

    ax.set_xlabel(name_o)
    ax.set_ylabel(name_m)
    ax.set_title(f'{name_m} by {name_o}')

    ax.grid()

def line_part_operator_intn(grr, ax, line_kws=None):
    """
    Interaction plot showing the part-operator interaction in a GRR study.

    Parameters
    ----------
    grr : mqr.msa.GRR
        Results from a GRR study.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    """
    line_kws = set_kws(
        line_kws,
        marker='_',
    )

    name_p = grr.names.part
    name_o = grr.names.operator
    name_m = grr.names.measurement

    intn = mqr.anova.interactions(
        grr.data,
        value=name_m,
        between=[name_p, name_o],
        formatted=False)

    ax.plot(intn, **line_kws)
    ax.set_xticks(intn.index)
    ax.set_xlabel(name_p)

    ax.grid()
    ax.legend(
        intn.columns,
        title=name_o,
        prop={'size': 8},
        fancybox=True,
        bbox_to_anchor=(1.02, 0.5, 0.0, 0.0),
        loc='center left',
        borderaxespad=0.0)

    ax.set_ylabel(f'{name_m}\n(mean over repeats)')
    ax.set_title(f'{name_p} * {name_o} Interaction')

def xbar_operator(grr, ax,
                  line_kws=None, text_kws=None,
                  target_kws=None, control_kws=None):
    """
    XBar-chart for the operator mean in a GRR study.

    Parameters
    ----------
    grr : mqr.msa.GRR
        Results from a GRR study.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    """
    target_kws = set_kws(
        target_kws,
        linewidth=0.5,
        color='k',
        zorder=0,
    )
    control_kws = set_kws(
        control_kws,
        linewidth=0.5,
        color='gray',
        zorder=0,
    )

    # Plot sample mean per operator
    name_p = grr.names.part
    name_o = grr.names.operator
    name_m = grr.names.measurement
    grp = grr.data.groupby([name_p, name_o])[name_m]
    mqr.plot.grouped_df(
        grp.mean().unstack(),
        ax=ax,
        line_kws=line_kws,
        text_kws=text_kws)

    # Add control bars
    if not np.all(grp.count() == grp.count().iloc[0]):
        raise ValueError('Only balanced experiments are supported.')

    N = grp.count().iloc[0]
    params = mqr.spc.XBarParams.from_range(
        np.mean(grr.data[name_m]),
        grp.apply(np.ptp).mean(),
        N,)

    ax.axhline(params.target(), **target_kws)
    ax.axhline(params.ucl(N), **control_kws)
    ax.axhline(params.lcl(N), **control_kws)

    ax.set_xlabel(name_p)
    ax.set_ylabel(f'{name_m}\n(mean over repeats)')
    ax.set_title(f'Xbar chart by {name_o}')

def r_operator(grr, ax,
               line_kws=None, text_kws=None,
               target_kws=None, control_kws=None):
    """
    R chart for the operator range in a GRR study.

    Parameters
    ----------
    grr : mqr.msa.GRR
        Results from a GRR study.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    """

    target_kws = set_kws(
        target_kws,
        linewidth=0.5,
        color='k',
        zorder=0,
    )
    control_kws = set_kws(
        control_kws,
        linewidth=0.5,
        color='gray',
        zorder=0,
    )

    # Plot sample range per operator
    name_p = grr.names.part
    name_o = grr.names.operator
    name_m = grr.names.measurement
    grp = grr.data.groupby([name_p, name_o])[name_m]
    range_r = grp.apply(np.ptp).unstack()
    mqr.plot.grouped_df(
        range_r,
        ax=ax,
        line_kws=line_kws,
        text_kws=text_kws)

    # Add control bars
    if not np.all(grp.count() == grp.count().iloc[0]):
        raise ValueError('Only balanced experiments are supported.')

    N = grp.count().iloc[0]
    rbar = np.mean(range_r)
    params = mqr.spc.RParams.from_range(rbar, N)

    ax.axhline(params.target(), **target_kws)
    ax.axhline(params.ucl(N), **control_kws)
    ax.axhline(params.lcl(N), **control_kws)
    ax.set_xlabel(name_p)
    ax.set_ylabel(f'{name_m}\n(mean over repeats)')
    ax.set_title(f'Range by {name_o}')

def grr(grr, axs, sources=None):
    """
    GRR summary plots.

    A 3 by 2 grid of:

    - bar graph of components of variation,
    - measurement by part,
    - R-chart by operator,
    - measurement by operator,
    - Xbar-chart by operator, and
    - part * operator interaction.

    This routine flattens the axes before drawing into them.

    Parameters
    ----------
    grr : mqr.msa.GRR
        GRR study.
    axs : numpy.ndarray
        A 3*2 array of matplotlib axes.
    sources : list[str], optional
        A list of components of variation to include in the bar graph (optional).

    Examples
    --------
    Create plots for a GRR analysis from the NIST silicon wafer resistivity.
    Data is from `<https://www.itl.nist.gov/div898/software/dataplot/data/MPC61.DAT>`_.

    Before creating the plot, though, there is a bit of data marshalling to do.
    This loads the data from the CSV file online into a DataFrame. The first 50
    rows are metadata etc., so skip those. Treat any one or more whitespace
    characters as a separator. And finally, add a column assigning numbers to
    repeated measurements, which allows the GRR routines to include repeats in
    the linear model.

    .. plot::
        :context: close-figs
        :nofigs:

        columns = ['RUNID', 'WAFERID', 'PROBE', 'MONTH', 'DAY', 'OPERATOR', 'TEMP', 'AVERAGE', 'STDDEV',]
        dtype = {
            'WAFERID': int,
            'PROBE':int,
        }

        data = pd.read_csv(
            'https://www.itl.nist.gov/div898/software/dataplot/data/MPC61.DAT',
            skiprows=50,
            header=None,
            names=columns,
            sep='\\\\s+',
            dtype=dtype,
            storage_options={'user-agent': 'github:nklsxn/mqr'}
        )
        data['REPEAT'] = np.repeat([1,2,3,4,5,6,7,8,9,10,11,12], 25)


    The GRR plots are created from a GRR study object :class:`mqr.msa.GRR`. For
    this example, use a tolerance of 8 ohm cm, and test for variance contribution
    from the probe (listed as operator). The name mapping shows the setup. This
    study uses only the first run `RUNID == 1`.

    .. plot::
        :context: close-figs

        tol = 2*8.0
        names = mqr.msa.NameMapping(
            part='WAFERID',
            operator='PROBE',
            measurement='AVERAGE')
        grr = mqr.msa.GRR(
            data.query('RUNID==1'),
            tolerance=tol,
            names=names,
            include_interaction=True)

        fig, axs = plt.subplots(3, 2, figsize=(10, 6), layout='constrained')
        mqr.plot.msa.grr(grr, axs=axs)


    """
    axs = axs.flatten()
    assert len(axs) == 6, 'GRR Tableau requires 6 subplot axes.'
    grr_table = mqr.msa.VarianceTable(grr)

    bar_var_pct(grr_table, sources=sources, ax=axs[0])
    box_measurement_by_part(grr, ax=axs[1])
    xbar_operator(grr, ax=axs[2])
    box_measurement_by_operator(grr, ax=axs[3])
    r_operator(grr, ax=axs[4])
    line_part_operator_intn(grr, ax=axs[5])



"""
Tool for producing plots in jupyter notebooks.
"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from mqr.plot.defaults import Defaults

class Figure(object):
    """
    Context manager that creates, shows and closes matplotlib subplots.

    Convenient in jupyter notebooks for reducing plotting boilerplate.

    Examples
    --------
    >>> with Figure(5, 3, m=1, n=2) as (fig, ax):
    >>>     ax[0].plot(...)
    >>>     seaborn.histplot(..., ax=ax[1])

    """

    def __init__(
            self,
            width=6, height=3,
            m=1, n=1,
            backend=None,
            filename=None,
            rc_params=None,
            gridspec_kw=None,
            subplot_kw=None,
            **fig_kw):
        """
        Construct the Figure context manager.

        Parameters
        ----------
        width, height : float, optional
            Width and height of the figure.
        m, n : int, optional
            Size of the grid of subplots. (Default 1 and 1.)
        backend : optional
            A matplotlib backend to switch to for the duration of the context
            manager. The previous backend is restored when the context manager
            exits. The default `None` makes no change to the backend.
        filename : str, optional
            A filename to save the figure to before closing the figure. The
            default `None` results in no change.
        gridspec_kw : dict, optional
            Passed as ``gridspec_kw=gridspec_kw` to :func:`matplotlib.pyplot.subplots`.
        subplot_kw : dict, optional
            Passed as ``subplot_kw=subplot_kw`` to :func:`matplotlib.pyplot.subplots`.
        fig_kw : kwargs, optional
            Keyword args passed as ``**fig_wk` to :func:`matplotlib.pyplot.subplots`.
        """
        fig_kw['figsize'] = (width, height)
        if 'layout' not in fig_kw:
            fig_kw['layout'] = 'constrained'
        self.m = m
        self.n = n
        self.prev_backend = None
        self.backend = backend
        self.filename = filename
        self.gridspec_kw = gridspec_kw
        self.subplot_kw = subplot_kw
        self.fig_kw = fig_kw

        if rc_params is None:
            rc_params = {}
        rc_params = Defaults.rc_params | rc_params
        self.rc_context = matplotlib.rc_context(rc_params)

    def __enter__(self):
        self.rc_context.__enter__()
        if self.backend is not None:
            self.prev_backend = matplotlib.get_backend()
            matplotlib.use(self.backend)
        fig, ax = plt.subplots(
            self.m, self.n,
            gridspec_kw=self.gridspec_kw,
            subplot_kw=self.subplot_kw,
            **self.fig_kw)
        self.fig = fig
        return fig, ax

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.filename is not None:
            plt.savefig(self.filename)
        plt.show()
        plt.close(self.fig)
        if self.prev_backend is not None:
            matplotlib.use(self.prev_backend)
        self.rc_context.__exit__(exc_type, exc_val, exc_tb)

def points_to_inches(width_pts, *, height_pts=None, hw_ratio=None):
    _PT_PER_INCH = 72.27
    width_inches = width_pts / _PT_PER_INCH
    if height_pts is not None:
        height_inches = height_pts / _PT_PER_INCH
    else:
        if hw_ratio is None:
            hw_ratio = 2 / (1 + np.sqrt(5))
        height_inches = width_inches * hw_ratio
    return width_inches, height_inches

def mm_to_inches(width_mm, *, height_mm=None, hw_ratio=None):
    _PT_PER_MM = 25.4
    width_inches = width_mm / _PT_PER_MM
    if height_mm is not None:
        height_inches = height_mm / _PT_PER_MM
    else:
        if hw_ratio is None:
            hw_ratio = 2 / (1 + np.sqrt(5))
        height_inches = width_inches * hw_ratio
    return width_inches, height_inches



"""
Plot confidence intervals.
"""

import matplotlib.pyplot as plt
import numpy as np

from mqr.inference.confint import ConfidenceInterval
from mqr.plot.defaults import Defaults
from mqr.plot.lib.util import set_kws

def confint(ci: ConfidenceInterval, ax, hyp_value=None, interval_kws=None, mean_kws=None, hyp_kws=None):
    """
    Draw a confidence interval with an hypothesised value; a graphical
    representation of an hypothesis test.

    When the the hypothesised value (`hyp_value`) is outside the confidence
    interval, the null-hypothesis is refejected at the confidence level used to
    construct the interval (`ci.conf`).

    Representing this relationship graphically makes sense when the distribution
    of the test statistic is the distribution of the sample. For example, in an
    hypothesis test for the mean of a sample, using a z-score as the test statistic.

    Parameters
    ----------
    ci : mqr.confint.ConfidenceInterval
        The confidence interval to draw.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    hyp_value : float, optional
        The null-hypothesis value.

    Examples
    --------
    Show a confidence interval on the mean.

    .. plot::
        :include-source:

        import mqr

        ci = mqr.inference.mean.confint_1sample(
            np.array([1, 5, 4, 3, 6, 2, 3, 3, 2, 5]),
            conf=0.98,
        )

        fig, ax = plt.subplots(figsize=(5, 2))
        mqr.plot.confint(ci, ax, hyp_value=3)
        plt.show()

    """
    interval_kws = set_kws(
        interval_kws,
        color='C0',
    )
    mean_kws = set_kws(
        mean_kws,
        marker=Defaults.marker,
        color='C0',
    )
    hyp_kws = set_kws(
        hyp_kws,
        marker=Defaults.marker,
        color='C1',
    )

    y_loc = 0

    l, r = ax.get_xlim()
    l_mkr = '<'
    r_mkr = '>'
    if np.isfinite(ci.lower):
        l = ci.lower
        l_mkr = '|'
    if np.isfinite(ci.upper):
        r = ci.upper
        r_mkr = '|'

    ax.plot(l, y_loc, marker=l_mkr, **interval_kws) # left marker
    ax.plot(r, y_loc, marker=r_mkr, **interval_kws) # right marker
    ax.plot([l, r], [y_loc, y_loc], **interval_kws) # interval line
    ax.plot(ci.value, y_loc, **mean_kws)
    if hyp_value is not None:
        ax.plot(hyp_value, y_loc, **hyp_kws)

    ax.set_yticks([y_loc])
    ax.set_yticklabels([ci.name])
    ax.tick_params(axis='y', left=False)
    ax.set_ylim(y_loc-1, y_loc+1)



"""
=================================================
Statistical process control (:mod:`mqr.plot.spc`)
=================================================

.. currentmodule:: mqr.plot.spc

.. rubric:: Functions
.. autosummary::
    :toctree: generated

    chart
    alarms
    oc
"""

import mqr
from mqr.plot.defaults import Defaults
from mqr.plot.lib.util import set_kws

import numpy as np
import scipy.stats as st

def chart(control_statistic, control_params, ax, *,
          line_kws=None, in_kws=None, out_kws=None, target_kws=None, control_kws=None):
    """
    Plots an control chart defined by ControlParams.

    Parameters
    ----------
    control_statistic : mqr.spc.ControlStatistic
        The control statistic from samples of a monitored process.
    control_params : mqr.spc.ControlParams
        The control parameters, usually calculated from reference/historical data.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    line_kws : dict, optional
        Keyword args for the statistic line (matplotlib.pyplot.plot).
    target_kws : dict, optional
        Keyword args for the target line (matplotlib.pyplot.axhline).
    control_kws : dict, optional
        Keyword args for the control lines (matplotlib.pyplot.plot).

    Examples
    --------
    This example charts the X-bar chart (sample mean) of a process with mean 1
    and standard deviation 5, using sample sizes of 6 observations. A total of
    20 samples are shown.

    .. plot::

        fig, ax = plt.subplots(figsize=(7, 3))

        # Raw data
        np.random.seed(0)
        x = pd.DataFrame(scipy.stats.norm(1, 5).rvs([20, 6]))

        # Parameters
        params = mqr.spc.XBarParams(centre=1, sigma=5)
        stat = params.statistic(x)

        # Charts
        mqr.plot.spc.chart(stat, params, ax=ax)

    """

    line_kws = set_kws(
        line_kws,
        marker=Defaults.marker,
        color='C0',
        zorder=1,
    )
    target_kws = set_kws(
        target_kws,
        linewidth=0.5,
        color='k',
        zorder=0,
    )
    control_kws = set_kws(
        control_kws,
        linewidth=0.5,
        color='gray',
        drawstyle='steps-mid',
        zorder=0,
    )

    stat = control_statistic.stat
    nobs = control_statistic.nobs
    index = stat.index
    target = control_params.target()
    lcl = control_params.lcl(nobs)
    ucl = control_params.ucl(nobs)

    line_values = [
        target,
        lcl.iloc[-1] if lcl is not None else None,
        ucl.iloc[-1] if ucl is not None else None,]
    yticks = [tick for tick in line_values if tick is not None]
    yticklabels = [
        f'{label}={value:g}'
        for label, value in zip(['target', 'LCL', 'UCL'], line_values)
        if value is not None]

    ax.plot(stat, **line_kws)
    ax.axhline(target, **target_kws)
    if lcl is not None:
        ax.plot(lcl, **control_kws)
    if ucl is not None:
        ax.plot(ucl, **control_kws)
    sec = ax.secondary_yaxis('right')
    sec.set_yticks(yticks)
    sec.set_yticklabels(yticklabels)
    ax.set_ymargin(0.15)
    ax.set_title(f'{control_params.name} chart')
    ax.set_xticks(index)

def alarms(control_statistic, control_params, control_rule, ax, *,
           point_kws=None, span_kws=None):
    """
    Plots alarms over a control chart.

    The control statistic and control parameters should be the same as the values
    passed to the `chart(...)` that drew the plot over which these alarms will be drawn.

    Alarm points are marked two ways:

    - a marker (a red dot by default) drawn over the statistic point, and
    - a region of colour (red by default) showing the span of alarmed points.

    Either can be switched off using point_kws/span_kws.

    Parameters
    ----------
    control_statistic : mqr.spc.ControlStatistic
        The control statistic from samples of a monitored process.
    control_params : mqr.spc.ControlParams
        The control parameters, usually calculated from reference/historical data.
    control_rule : Callable[[ControlStatistic, ControlParams], pandas.Series[bool]]
        A map from the control statistic and control parameters to a series of
        bools with a corresponding index, where True values are an alarm at that 
        index.
    ax : matplotlib.axes.Axes
    point_kws : dict, optional
        Keyword args for the alarm markers (matplotlib.pyplot.plot).
    span_kws : dict, optional
        Keyword args for the alarm regions (matplotlib.pyplot.axvspan).

    Examples
    --------
    This example shows an X-bar chart (sample mean) of a process with mean 1
    and standard deviation 5, using sample sizes of 6 observations. A total of
    20 samples are shown. The limits are 2 standard deviations of the standard
    error of the mean.

    .. plot::

        fig, ax = plt.subplots(figsize=(7, 3))

        # Raw data
        np.random.seed(0)
        x = pd.DataFrame(
            scipy.stats.norm(1, 5).rvs([20, 6]),
            columns=range(6))

        # Parameters
        params = mqr.spc.XBarParams(centre=1, sigma=5, nsigma=2)
        stat = params.statistic(x)
        rule = mqr.spc.rules.limits()

        # Charts
        mqr.plot.spc.chart(stat, params, ax=ax)
        mqr.plot.spc.alarms(stat, params, rule, ax=ax)

    """

    point_kws = set_kws(
        point_kws,
        linewidth=0,
        color='C3',
        marker='.',
    )
    span_kws = set_kws(
        span_kws,
        color='C3',
        alpha=0.2,
        zorder=-1,
    )

    stat = control_statistic.stat
    alarms = control_rule(control_statistic, control_params)
    for alarm in mqr.spc.util.alarm_subsets(alarms):
        ax.plot(stat.loc[alarm], **point_kws)
        a = alarm[0] - 0.5
        b = alarm[-1] + 0.5
        ax.axvspan(a, b, **span_kws)

def oc(n, c, ax, defect_range=None, line_kws=None):
    """
    Plot an OC curve.

    Parameters
    ----------
    n : int
        Sample size.
    c : int
        Acceptance number.
    ax : matplotlib.axes.Axes
        Axes for the plot.
    defect_range : tuple[float, float], optional
        Range of defect rates to show (on the x-axis).
    line_kws : dict, optional
        Keyword args for the line. Passed to :func:`matplotlib.pyplot.plot`.

    Examples
    --------
    This example shows an OC curve with sample size 40 and acceptance number 6.
    The plot shows only defect rates between 0 to 0.3 (along the x axis).

    .. plot::

        fig, ax = plt.subplots(figsize=(7, 3))

        mqr.plot.spc.oc(n=40, c=6, defect_range=(0, 0.3), ax=ax)

    """
    line_kws = set_kws(
        line_kws,
        color='C0',
    )

    if defect_range is None:
        defect_range = (0, 1)

    ps = np.linspace(*defect_range)

    pa = np.empty(ps.shape)
    for i in range(len(ps)):
        pa[i] = st.binom(n, ps[i]).cdf(c)

    ax.plot(ps, pa, **line_kws)
    ax.grid()
    ax.set_xlabel('Defect rate')
    ax.set_ylabel('Prob of Acceptance')



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import importlib

project = 'mqr'
copyright = '2024-2025 Nikolas Crossan and Kevin Otto'
author = 'Nikolas Crossan'
version = importlib.metadata.version('mqrpy')
release = version


# -- App setup ---------------------------------------------------------------

def setup(app):
    app.add_css_file('_static/mqr.css')


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.duration',
    'sphinx.ext.githubpages',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx_rtd_theme',
    'numpydoc',
    'matplotlib.sphinxext.plot_directive',

    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',

    # 'myst_parser',
    'myst_nb',
]

autosummary_generate = True

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
add_function_parentheses = False


# -- Options for myst-parser and myst-nb -------------------------------------
# https://myst-parser.readthedocs.io/en/latest/syntax/optional.html

source_suffix = {
    '.rst': 'restructuredtext',
    '.ipynb': 'myst-nb',
    '.myst': 'myst-nb',
}

myst_enable_extensions = [
    "amsmath", #
    "attrs_block", #
    "attrs_inline", #
    "colon_fence", #
    "deflist", #
    "dollarmath", #
    "fieldlist", #
    "html_admonition", #
    # "html_image",
    # "linkify",
    "replacements", #
    "smartquotes", #
    # "strikethrough",
    # "substitution",
    "tasklist", #
]
myst_footnote_transition = False
myst_heading_anchors = 3
myst_links_external_new_tab = True
myst_url_schemes = {
    "doi": "https://doi.org/{{path}}",
    "http": None,
    "https": None,
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_sidebars = {
    "index": ["search-button-field"],
    "**": ["search-button-field", "sidebar-nav-bs"]
}
html_theme_options = {
    "github_url": "https://github.com/nklsxn/mqr",
    "header_links_before_dropdown": 6,
    "icon_links": [],
    "logo": {
        "text": "MQR",
    },
    "navbar_start": ["navbar-logo"],
    "navbar_end": ["version-switcher", "theme-switcher", "navbar-icon-links"],
    "navbar_persistent": [],
    "show_version_warning_banner": True,
    "secondary_sidebar_items": ["page-toc"],
    "switcher": {
        "version_match": version,
        "json_url": "https://raw.githubusercontent.com/nklsxn/mqr/refs/heads/master/docs/switcher.json"
    }
}
html_context = {
    "default_mode": "light",
}

# html_title = f"{project} v{version} Manual"
# html_last_updated_fmt = "%b %d, %Y"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_css_files = [
    'mqr.css',
]
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = "project-templatedoc"

# mathjax_path = "scipy-mathjax/MathJax.js?config=scipy-mathjax"


# -- Intersphinx configuration -------------------------------------------------

intersphinx_mapping = {
    'matplotlib': ('https://matplotlib.org/stable', None),
    'seaborn': ('https://seaborn.pydata.org/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable', None),
    'scipy': ('https://docs.scipy.org/doc/scipy', None),
    'statsmodels': ('https://www.statsmodels.org/stable', None),
    'reliability': ('https://reliability.readthedocs.io/en/latest', None),
    'pyDOE3': ('https://pydoe3.readthedocs.io/en/latest', None),
}


# -- matplotlib configuration --------------------------------------------------

import matplotlib as mpl
import mqr

plot_include_source = True
plot_pre_code = '''
import numpy as np
import pandas as pd
import scipy
from matplotlib import pyplot as plt
import mqr
'''
plot_formats = ['png', 'pdf']
plot_rcparams = mqr.plot.defaults.Defaults.rc_params
plot_apply_rcparams = True



import mqr

import numpy as np
import pandas as pd
import pytest


################################################################################
## Fixtures

@pytest.fixture
def sample1():
    data = np.array([
        [1, 2, 3, 1, 2, 3],
        [2, 4, 6, 6, 4, 2],
        [1, 2, 0, 2, 0, 1],
    ])
    return pd.DataFrame(
        data,
        index=range(1, 4),
        columns=[f'x{i}' for i in range(1, 7)])

@pytest.fixture
def sample2():
    data = np.array([
        [1, 2],
        [2, 4],
        [3, 6],
    ])
    return pd.DataFrame(
        data,
        index=range(1, 4),
        columns=[f'x{i}' for i in range(1, 3)])

@pytest.fixture
def sample3():
    data = np.array([
        [0, 2],
        [0, 2],
        [3, 5],
        [3, 5],
        [3, 5],
        [3, 5],
        [3, 5],
        [3, 5],
    ])
    m, n = data.shape
    return pd.DataFrame(
        data,
        index=range(1, m+1),
        columns=[f'' for i in range(1, n+1)])

@pytest.fixture
def sample4():
    data = np.array([
        [1, 5, 5],
        [1, 5, 5],
        [1, 10, 6],
        [1, 10, 5],
        [1, 5, 5],
        [1, 5, 5],
        [4, 10, 6],
        [1, 10, 5],
        [1, 5, 5],
        [1, 5, 5],
    ])
    m, n = data.shape
    return pd.DataFrame(
        data,
        index=range(1, m+1),
        columns=[f'x{i}' for i in range(1, n+1)])

# Same as sample4 but index starts at 5
@pytest.fixture
def sample5():
    data = np.array([
        [1, 5, 5],
        [1, 5, 5],
        [1, 10, 6],
        [1, 10, 5],
        [1, 5, 5],
        [1, 5, 5],
        [4, 10, 6],
        [1, 10, 5],
        [1, 5, 5],
        [1, 5, 5],
    ])
    m, n = data.shape
    return pd.DataFrame(
        data,
        index=range(5, m+5),
        columns=[f'x{i}' for i in range(1, n+1)])

################################################################################
## Tests

def test_ControlStatistic_post_init():
    stat = pd.Series([1, 2, 3, 4, 5])
    nobs = pd.Series([4, 4, 4, 4])
    with pytest.raises(ValueError):
        mqr.spc.ControlStatistic(stat, nobs)


## XBarParams

def test_XBarParams_statistic(sample1):
    params = mqr.spc.XBarParams(0, 1)
    stat = params.statistic(sample1)

    assert list(stat.stat) == [2, 4, 1]
    assert list(stat.nobs) == [6, 6, 6]

def test_XBarParams_se():
    params = mqr.spc.XBarParams(0, 4)
    assert params.se(16) == 1

def test_XBarParams_target():
    params = mqr.spc.XBarParams(3, 1)
    assert params.target() == 3

def test_XBarParams_lcl():
    params = mqr.spc.XBarParams(7, 6, 5)
    assert params.lcl(9) == 7 - (6 / 3) * 5

def test_XBarParams_ucl():
    params = mqr.spc.XBarParams(7, 6, 5)
    assert params.ucl(9) == 7 + (6 / 3) * 5

def test_XBarParams_from_stddev():
    params = mqr.spc.XBarParams.from_stddev(11, 7, 5, 13)
    assert params.centre == 11
    assert params.sigma == 7 / mqr.spc.util.c4(5)
    assert params.nsigma == 13

def test_XBarParams_from_range():
    params = mqr.spc.XBarParams.from_range(11, 7, 5, 13)
    assert params.centre == 11
    assert params.sigma == 7 / mqr.spc.util.d2(5)
    assert params.nsigma == 13

def test_XBarParams_from_data(sample1, sample2):
    params_sbar = mqr.spc.XBarParams.from_data(sample2, 's_bar', 5)
    params_rbar = mqr.spc.XBarParams.from_data(sample2, 'r_bar', 5)

    assert params_sbar.centre == 3
    assert params_rbar.centre == 3

    assert params_sbar.sigma == np.sqrt(2) / mqr.spc.util.c4(2)
    assert params_rbar.sigma == 2 / mqr.spc.util.d2(2)

    assert params_sbar.nsigma == 5
    assert params_rbar.nsigma == 5

    params_sbar = mqr.spc.XBarParams.from_data(sample1, 's_bar', 5)
    params_rbar = mqr.spc.XBarParams.from_data(sample1, 'r_bar', 5)

    assert params_sbar.centre == 7 / 3
    assert params_rbar.centre == 7 / 3

    assert params_sbar.sigma == np.mean(np.std(sample1, ddof=1, axis=1)) / mqr.spc.util.c4(6)
    assert params_rbar.sigma == np.mean(np.ptp(sample1, axis=1)) / mqr.spc.util.d2(6)

    assert params_sbar.nsigma == 5
    assert params_rbar.nsigma == 5


## RParams

def test_RParams_statistic(sample1):
    params = mqr.spc.RParams(3, 5, 7)
    stat = params.statistic(sample1)
    assert list(stat.stat) == [2, 4, 2]
    assert list(stat.nobs) == [6, 6, 6]

def test_RParams_se():
    params = mqr.spc.RParams(3, 5, 7)
    assert params.se(13) == mqr.spc.util.d3(13) * 5

def test_RParams_target():
    params = mqr.spc.RParams(3, 5, 7)
    assert params.target() == 3

def test_RParams_lcl():
    params = mqr.spc.RParams(11, 7, 1)
    assert params.lcl(13) == 11 - mqr.spc.util.d3(13) * 7

    params = mqr.spc.RParams(11, 7, 5)
    assert params.lcl(13) == 0

def test_RParams_ucl():
    params = mqr.spc.RParams(11, 7, 5)
    assert params.ucl(13) == pytest.approx(11 + 5 * mqr.spc.util.d3(13) * 7, 1e-15)

def test_RParams_from_range():
    params = mqr.spc.RParams.from_range(3, 5, 7)
    assert params.centre == 3
    assert params.sigma == 3 / mqr.spc.util.d2(5)
    assert params.nsigma == 7

def test_RParams_from_data(sample1):
    params = mqr.spc.RParams.from_data(sample1, 5)
    assert params.centre == np.mean(np.ptp(sample1, axis=1))
    assert params.sigma == params.centre / mqr.spc.util.d2(6)

## SParams

def test_SParams_statistic(sample1):
    params = mqr.spc.SParams(3, 5, 7)
    stat = params.statistic(sample1)
    assert list(stat.stat) == list(np.std(sample1, ddof=1, axis=1))
    assert list(stat.nobs) == [6, 6, 6]

def test_SParams_se():
    params = mqr.spc.SParams(3, 5, 7)
    c4 = mqr.spc.util.c4(6)
    assert params.se(6) == 3 * np.sqrt(1 - c4**2) / c4

def test_SParams_target():
    params = mqr.spc.SParams(3, 5, 7)
    assert params.target() == 3

def test_SParams_lcl():
    params = mqr.spc.SParams(3, 1)
    c4 = mqr.spc.util.c4(6)
    assert params.lcl(6) == 3 - 3 * np.sqrt(1 - c4**2) / c4

    params = mqr.spc.SParams(3, 5)
    assert params.lcl(6) == 0

def test_SParams_ucl():
    params = mqr.spc.SParams(3, 5)
    c4 = mqr.spc.util.c4(6)
    assert params.ucl(6) == 3 + 3 * 5 * np.sqrt(1 - c4**2) / c4

def test_SParams_from_data(sample1):
    params = mqr.spc.SParams.from_data(sample1, 5)
    assert params.centre == np.mean(np.std(sample1, ddof=1, axis=1))
    assert params.nsigma == 5

## EwmaParams

def test_EwmaParams_statistic(sample3):
    params = mqr.spc.EwmaParams(1, 1, 0.2, 3)
    # The strategy prepends the target as the first value, which is 1 here.
    stat = params.statistic(sample3.iloc[1:])
    # The 1 in the test sample is used as the initial value in the series,
    # then drop the first value (the initialisation value) because only the
    # sample values are returned by the strategy.
    assert list(stat.stat) == list(sample3.mean(axis=1).ewm(alpha=0.2, adjust=False).mean().iloc[1:])
    assert list(stat.nobs) == [2, 2, 2, 2, 2, 2, 2]

def test_EwmaParams_target():
    params = mqr.spc.EwmaParams(5, 1, 0.2, 3)
    assert params.target() == 5

def test_EwmaParams_lcl():
    """
    Check against equations in Chapter 9 of [1]_.

    References
    ----------
    .. [1]  Montgomery, D. C. (2009).
            Statistical quality control (Vol. 7).
            New York: Wiley.
    """
    i = np.array([3, 4, 5, 6])
    nobs = pd.Series([9, 9, 9, 9], index=i)
    frac = 0.2 / (2 - 0.2)
    decay = (1 - (1 - 0.2)**(2 * i))

    params = mqr.spc.EwmaParams(1, 6, 0.2, 5, False)
    assert list(params.lcl(nobs)) == list(1 - 5 * 6/3 * np.sqrt(frac * decay))

    params = mqr.spc.EwmaParams(1, 6, 0.2, 5, True)
    assert all(params.lcl(nobs) == 1 - 5 * 6/3 * np.sqrt(frac))

def test_EwmaParams_ucl():
    """
    Check against equations in Chapter 9 of [1]_.

    References
    ----------
    .. [1]  Montgomery, D. C. (2009).
            Statistical quality control (Vol. 7).
            New York: Wiley.
    """
    i = np.array([3, 4, 5, 6])
    nobs = pd.Series([9, 9, 9, 9], index=i)
    frac = 0.2 / (2 - 0.2)
    decay = (1 - (1 - 0.2)**(2 * i))

    params = mqr.spc.EwmaParams(1, 6, 0.2, 5, False)
    assert list(params.ucl(nobs)) == list(1 + 5 * 6/3 * np.sqrt(frac * decay))

    params = mqr.spc.EwmaParams(1, 6, 0.2, 5, True)
    assert all(params.ucl(nobs) == 1 + 5 * 6/3 * np.sqrt(frac))

def test_EwmaParams_from_stddev():
    params = mqr.spc.EwmaParams.from_stddev(1.2, 3.4, 5, 0.6, 7, False)
    assert params.mu_0 == 1.2
    assert params.sigma == 3.4 / mqr.spc.util.c4(5)
    assert params.lmda == 0.6
    assert params.L == 7
    assert not params.steady_state

    params = mqr.spc.EwmaParams.from_stddev(1.2, 3.4, 5, 0.6, 7, True)
    assert params.mu_0 == 1.2
    assert params.sigma == 3.4 / mqr.spc.util.c4(5)
    assert params.lmda == 0.6
    assert params.L == 7
    assert params.steady_state

def test_EwmaParams_from_data(sample1):
    params = mqr.spc.EwmaParams.from_data(sample1, 0.3, 4, True)
    target = np.mean(sample1)
    sigma = np.mean(np.std(sample1, ddof=1, axis=1)) / mqr.spc.util.c4(6)
    assert params.mu_0 == target
    assert params.sigma == sigma
    assert params.lmda == 0.3
    assert params.L == 4
    assert params.steady_state

    params = mqr.spc.EwmaParams.from_data(sample1, 0.3, 4, False)
    target = np.mean(sample1)
    sigma = np.mean(np.std(sample1, ddof=1, axis=1)) / mqr.spc.util.c4(6)
    assert params.mu_0 == target
    assert params.sigma == sigma
    assert params.lmda == 0.3
    assert params.L == 4
    assert not params.steady_state

## MewmaParams

def test_MewmaParams_statistic(sample5):
    """
    Check against manual construction of EWMA and statistic
    """
    sample = sample5
    idx = sample.index
    lmda = 0.2
    params = mqr.spc.MewmaParams.from_data(sample, np.nan, lmda)
    stat = params.statistic(sample)

    z = pd.DataFrame(index=sample.index, columns=sample.columns)
    z.loc[idx[0]] = lmda * (sample.loc[idx[0]] - params.mu)
    for i in idx[1:]:
        z.loc[i] = lmda * (sample.loc[i] - params.mu) + (1 - lmda) * z.loc[i-1]

    t2 = pd.Series(index=sample.index)
    for i in idx:
        cov_scaled = lmda / (2 - lmda) * params.cov  * (1 - (1 - lmda)**(2*i))
        t2.loc[i] = z.loc[i] @ np.linalg.inv(cov_scaled) @ z.loc[i]

    assert list(stat.stat.index) == list(sample.index)
    assert list(stat.nobs.index) == list(sample.index)
    assert np.all(np.isclose(stat.stat, t2))
    assert all(stat.nobs == 3)

def test_MewmaParams_target(sample4):
    params = mqr.spc.MewmaParams.from_data(sample4, 3.45, 0.5)
    assert params.target() == 0

def test_MewmaParams_lcl(sample4):
    params = mqr.spc.MewmaParams.from_data(sample4, 3.45, 0.2)
    assert params.lcl(sample4) == None

def test_MewmaParams_ucl(sample4):
    params = mqr.spc.MewmaParams.from_data(sample4, 3.45, 0.2)
    ucl = params.ucl(sample4)
    assert list(ucl.index) == list(sample4.index)
    assert all(ucl == 3.45)

def test_MewmaParams_from_data(sample4):
    params = mqr.spc.MewmaParams.from_data(sample4, 3.45, 0.2)
    assert list(params.mu) == list(np.mean(sample4, axis=0))
    assert np.all(np.isclose(params.cov, np.cov(sample4.T, ddof=1)))
    assert params.lmda == 0.2
    assert params.limit == 3.45



from mqr.inference.hyptest import HypothesisTest

import mqr.inference.lib.util as util

import scipy

def test_nsample(*x, alternative='two-sided', method='levene'):
    """
    Hypothesis test for homogeneity of variances of multiple samples.

    Null hypothesis
        var(`x[i]`) == var(`x[j]`), for all `i`, `j`

    Parameters
    ----------
    x : array_like
        Test equality of variances of these samples.
    alternative : {'two-sided'}, optional
        Sense of alternative hypothesis. Only 'two-sided' is currently supported.
    method : {'levene', 'fligner-killeen'}, optional
        | 'levene'
        |   Levene's test for homogeneity of variance.
            Calls :func:`scipy..levene <scipy.stats.levene>`.
        | 'fligner-killeen'
        |   Fligner-Killeen test for homogeneity of variance.
            Calls :func:`scipy..fligner <scipy.stats.fligner>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    if method == 'levene':
        if alternative != 'two-sided':
            raise ValueError('One-sided alternative not available in Levene test.')
        (stat, pvalue) = scipy.stats.levene(*x)
    elif method == 'fligner-killeen':
        if alternative != 'two-sided':
            raise ValueError('One-sided alternative not available in Fligner-Killeen test.')
        (stat, pvalue) = scipy.stats.fligner(*x)
    else:
        raise ValueError(util.method_error_msg(method, ['levene', 'fligner-killeen']))

    return HypothesisTest(
        description='equality of variances',
        alternative=alternative,
        method=method,
        sample_stat='var(x_i)',
        sample_stat_target='var(x_j)',
        sample_stat_value=stat,
        stat=stat,
        pvalue=pvalue,)



"""
SPC Types

.. :currentmodule:`mqr.spc`

This module has types for creating control parameters and statistics.
"""

import abc
from dataclasses import asdict, dataclass, field
import matplotlib.pyplot as plt
from mqr.spc.util import c4, d2, d3
import numpy as np
import pandas as pd
import scipy

@dataclass
class ControlStatistic:
    """
    Monitored statistic of a controlled process.

    Parameters
    ----------
    stat: pandas.Series
        Monitored statistic, indexed by sample.
    nobs: pandas.Series
        Number of observations used to construct the corresponding sample.
    """
    stat: pd.Series = field(repr=False)
    nobs: pd.Series = field(repr=False)

    def __post_init__(self):
        if len(self.stat) != len(self.nobs):
            raise ValueError('Series stat and nobs must be the same length.')

@dataclass
class ControlParams:
    """
    Base class of all control methods.

    Other types should implement at least these functions to be monitored with
    SPC rules.
    """

    @abc.abstractmethod
    def statistic(self, samples):
        """
        Calculates the control statistic from samples.

        Parameters
        ----------
        samples : pandas.DataFrame
            Observations to be controlled. The index is a sample number, which
            some methods (eg. EWMA) depend on. The columns are dependent on the
            monitoring method, for example, they could be the observations in a
            fixed-size sample.

        Returns
        -------
        :class:`ControlStatistic`
            Statistics and sample sizes for this method, based on samples. Has
            the same index as the argument `samples`.
        """
        pass

    @abc.abstractmethod
    def target(self):
        """
        Process target value for this monitoring method.

        Returns
        -------
        float
            Target value.
        """
        pass

    @abc.abstractmethod
    def lcl(self, nobs):
        """
        Calculates the lower control limits for samples with sizes in `nobs`.

        Parameters
        ----------
        nobs : pandas.Series
            Sample sizes used to construct the statistic that will be monitored
            with this limit. The index must match the index of the `samples`
            argument passed to `ControlParams.statistic`.

        Returns
        -------
        pandas.Series
            Lower limits for each sample. Has the same index as the argument `nobs`.
        """
        pass

    @abc.abstractmethod
    def ucl(self, nobs):
        """
        Calculates the upper control limits for samples with sizes in `nobs`.

        Parameters
        ----------
        nobs : pandas.Series
            Sample sizes used to construct the statistic that will be monitored
            with this limit. The index must match the index of the `samples`
            argument passed to `ControlParams.statistic`.

        Returns
        -------
        pandas.Series
            Upper limits for each sample. Has the same index as the argument `nobs`.
        """
        pass

    def asdict(self):
        """
        Returns a dict representation of this class.

        This is a convenience method for serialisation. It allows users to
        store control parameters that are based on historical data, ready for
        later use. The resulting dictionary can be passed to the class's
        constructor to reproduce an identical object.
        """
        return asdict(self)

@dataclass
class ShewhartParams(ControlParams):
    """
    Base class of Shewhart-type control methods.

    Other types should implement at least these functions to be monitored with
    SPC rules expressed in terms of "sigma"s, or standard deviations of the
    sampling distribution used by the control method.
    """
    @abc.abstractmethod
    def se(self, nobs):
        """
        Calculates the standard error of samples with sizes `nobs`.

        Parameters
        ----------
        nobs : pandas.Series
            Sample sizes used to construct the statistic that will be monitored
            with this limit. The index must match the index of the `samples`
            argument passed to `ControlParams.statistic`.

        Notes
        -----
        This function calculates the standard deviation of the sampling
        distribution for the statistic that is being monitored. For example,
        in an XBar chart, `se` returns the standard deviation of the sampling
        distribution of the mean, ie. the standard error of the mean.

        Returns
        -------
        pandas.Series
            Magnitude of one standard deviation of the sampling distribution of
            this statistic for each sample size in `nobs`. Has the same index as
            the argument `nobs`.
        """
        pass

@dataclass
class XBarParams(ShewhartParams):
    """
    XBar (sample mean) control parameters.

    Use this class to construct traditional XBar charts. In addition to
    construction using class attributes, `XBarParams` can be constructed from
    sample standard deviation, sample range or historical data.

    Attributes
    ----------
    centre : float
        Centreline of the process distribution.
    sigma : float
        Standard deviation of the process distribution.
    nsigma : float, optional
        Distance in standard errors of the upper and lower control limits from
        the centreline.
    name : str, optional
        Name of the control method.
    """
    centre: float = field(repr=True)
    sigma: float = field(repr=True)
    nsigma: float = field(default=3, repr=False)

    name: str = field(default='XBar', repr=False)

    def statistic(self, samples):
        """
        XBar statistic; the sample mean.

        Parameters
        ----------
        samples : pandas.DataFrame
            Samples indexed by sample ID/number with columns for each observation.
            See :meth:`mqr.spc.ControlParams.statistic`.

        Returns
        -------
        :class:`ControlStatistic`
            Means of `samples`.
        """
        return ControlStatistic(
            stat=samples.mean(axis=1),
            nobs=samples.apply(len, axis=1))

    def se(self, nobs):
        """
        Standard error of xbar (standard error of the mean).

        Parameters
        ----------
        nobs : pandas.Series
            Number of observations in each sample. Index of `nobs` corresponds
            to the index of the argument `samples` passed to `statistic`.

        Returns
        -------
        pandas.Series
            Standard error of the mean for samples with sizes `nobs`.
        """
        return self.sigma / np.sqrt(nobs)

    def target(self):
        """
        Expected value of the sample mean.

        Returns
        -------
        float
        """
        return self.centre

    def lcl(self, nobs):
        """
        Calculates the lower control limits for samples with sizes in `nobs`.

        Parameters
        ----------
        nobs : pandas.Series
            Number of observations in each sample. Index of `nobs` corresponds
            to the index of the argument `samples` passed to `statistic`.

        Returns
        -------
        pandas.Series
        """
        return self.centre - self.nsigma * self.se(nobs)

    def ucl(self, nobs):
        """
        Calculates the upper control limits for samples with sizes in `nobs`.

        Parameters
        ----------
        nobs : pandas.Series
            Number of observations in each sample. Index of `nobs` corresponds
            to the index of the argument `samples` passed to `statistic`.

        Returns
        -------
        pandas.Series
        """
        return self.centre + self.nsigma * self.se(nobs)

    @staticmethod
    def from_stddev(centre, s_bar, nobs, nsigma=3):
        """
        Constructs XBarParams from an average sample stddev.

        Parameters
        ----------
        centre : float
            Process target mean.
        s_bar : float
            Average sample stddev from a reference in-control process.
        nobs : int
            Size (fixed) of samples used to calculate (s_bar).
        nsigma : float
            Number of stderr to set control limits.
        """
        return XBarParams(centre, s_bar / c4(nobs), nsigma, 'XBar(S)')

    @staticmethod
    def from_range(centre, r_bar, nobs, nsigma=3):
        """
        Constructs XBarParams from an average range.

        Using sample range to estimate process standard deviation can be
        inefficient. See :doc:`/notes/stddev-efficiency`.

        Parameters
        ----------
        centre : float
            Process target mean.
        r_bar : float
            Average sample range from a reference in-control process.
        nobs : int
            Size (fixed) of samples used to calculate (`r_bar`).
        nsigma : float
            Number of stderr to set control limits.
        """
        return XBarParams(centre, r_bar / d2(nobs), nsigma, 'XBar(R)')

    @staticmethod
    def from_data(samples, method='s_bar', nsigma=3):
        """
        Constructs XBarParams from reference samples.

        Reference samples are usually taken from an historical, in-control process.

        .. note::
            Using sample range to estimate process standard deviation can be
            inefficient. See :doc:`/notes/stddev-efficiency`.

        Parameters
        ----------
        samples : pandas.DataFrame
            Reference samples. Rows represent samples and columns represent
            observations.
        method : {'s_bar', 'r_bar'}, optional
            | 's_bar'
            |   Estimates process standard deviation from the sample standard
                deviation. Produces the traditional XBar-S chart.
            | 'r_bar'
            |   Estimates process standard deviation from the sample range.
                Produces the traditional XBar-R chart.
                See notes about efficiency :doc:`/notes/stddev-efficiency`.
        nsigma : float, optional
            Number of stderr to set control limits.
        """
        if method == 's_bar':
            centre = np.mean(samples, axis=1).mean()
            s_bar = np.std(samples.values, ddof=1, axis=1).mean()
            nobs = samples.shape[1]
            return XBarParams.from_stddev(centre, s_bar, nobs, nsigma)
        elif method == 'r_bar':
            centre = samples.mean(axis=1).mean()
            r_bar = np.ptp(samples, axis=1).mean()
            nobs = samples.shape[1]
            return XBarParams.from_range(centre, r_bar, nobs, nsigma)
        else:
            raise ValueError(f'Method {method} not supported.')

@dataclass
class RParams(ShewhartParams):
    """
    R (sample range) control parameters.

    Use this class to construct traditional R charts. In addition to
    construction using class attributes, `RParams` can be constructed from
    sample range or historical data.

    .. note::
        Using sample range to estimate process standard deviation can be
        inefficient. See :doc:`/notes/stddev-efficiency`.

    Attributes
    ----------
    centre : float
        Expected value of the process range.
    sigma : float
        Standard deviation of the process.
    nsigma : float, optional
        Distance in standard errors (of the range) of the upper and lower
        control limits from the centreline.
    name : str, optional
        Name of the control method.
    """
    centre: float = field(repr=True)
    sigma: float = field(repr=True)
    nsigma: float = field(repr=False, default=3)

    name: str = field(default='R', repr=False)

    def statistic(self, samples):
        """
        R statistic; the sample range.
        """
        return ControlStatistic(
            stat=np.ptp(samples, axis=1),
            nobs=samples.apply(len, axis=1))

    def se(self, nobs):
        """
        Standard error of the range.
        """
        return d3(nobs) * self.sigma

    def target(self):
        """
        Expected value of the sample range.
        """
        return self.centre

    def lcl(self, nobs):
        """
        Calculates the lower control limits for samples with sizes in `nobs`.
        """
        return np.clip(self.centre - self.nsigma * self.se(nobs), 0, np.inf)

    def ucl(self, nobs):
        """
        Calculates the upper control limits for samples with sizes in `nobs`.
        """
        return self.centre + self.nsigma * self.se(nobs)

    @staticmethod
    def from_range(r_bar, nobs, nsigma=3):
        """
        Constructs RParams from process range.

        Parameters
        ----------
        r_bar : float
            Expected value of process range.
        nobs : int
            Sample size used to estimate r_bar.
        nsigma : float
            Number of stderr to set control limits.
        """
        return RParams(r_bar, r_bar / d2(nobs), nsigma)

    @staticmethod
    def from_data(samples, nsigma=3):
        """
        Constructs RParams from reference samples.

        Reference samples are usually taken from an historical, in-control process.

        Parameters
        ----------
        samples : pandas.DataFrame
            Reference samples.
        nsigma : float
            Number of stderr to set control limits.
        """
        r_bar = np.ptp(samples, axis=1).mean()
        nobs = samples.shape[1]
        return RParams(r_bar, r_bar / d2(nobs), nsigma)

@dataclass
class SParams(ShewhartParams):
    """
    S (sample standard deviation) control parameters.

    Use this class to construct traditional S charts. In addition to
    construction using class attributes, `SParams` can be constructed from
    historical data.

    Attributes
    ----------
    centre : float
        Expected value of the process stddev.
    nsigma : float, optional
        Distance in standard errors (of the stddev) of the upper and lower
        control limits from the centreline.
    name : str, optional
        Name of the control method.
    """
    centre: float = field(repr=False)
    nsigma: int = field(repr=False, default=3)

    name: str = field(default='S', repr=False)

    def statistic(self, samples):
        """
        S statistic; the sample standard deviation.
        """
        return ControlStatistic(
            stat=samples.std(axis=1, ddof=1),
            nobs=samples.apply(len, axis=1))

    def se(self, nobs):
        """
        Standard error of the standard deviation
        """
        return self.centre * np.sqrt(1 - c4(nobs)**2) / c4(nobs)

    def target(self):
        """
        Expected value of the standard deviation.
        """
        return self.centre

    def lcl(self, nobs):
        """
        Calculates the lower control limits for samples with sizes in `nobs`.
        """
        return np.clip(self.target() - self.nsigma * self.se(nobs), 0, np.inf)

    def ucl(self, nobs):
        """
        Calculates the upper control limits for samples with sizes in `nobs`.
        """
        return self.target() + self.nsigma * self.se(nobs)

    @staticmethod
    def from_data(samples, nsigma=3):
        """
        Constructs SParams from reference samples.

        Reference samples are usually taken from an historical, in-control process.

        Parameters
        ----------
        samples : pandas.DataFrame
            Reference samples.
        nsigma : float
            Number of stderr to set control limits.
        """
        nobs = samples.shape[1]
        centre = samples.std(axis=1, ddof=1).mean()
        return SParams(centre, nsigma)

@dataclass
class EwmaParams(ControlParams):
    """
    EWMA control parameters.

    Use this class to construct EWMA charts for the mean of a sample. In addition
    to construction using class attributes, `EwmaParams` can be constructed from
    historical data.

    Attributes
    ----------
    mu_o : float
        Desired process mean.
    sigma : float
        Process standard deviation.
    lmda : float
        Decay rate.
    L : float
        Width of the control limits, in multiples of the standard deviation of
        the smoothed mean.
    steady_state : bool, optional
        Whether the process has already decayed to the steady state control limits.
    name : str, optional
        Name of the control method.

    Notes
    -----
    This monitoring strategy tracks a statistic based on either an observation
    or a sample. When the sample size is 1, the statistic is the EWMA of those
    samples. When the sample size is > 1, the statistic is the EWMA of the mean
    of those samples.

    The parameter sigma is an estimate of the standard deviation of the process,
    but the control limits are calculated from the standard error of the mean.
    When the sample size is 1, the standard error of the mean equals the standard
    deviation of the process.

    When the `EwmaParams` are created from data, `nobs` in the historical data
    need not match the sample size used to create the tracked statistic. The
    control limits will be scaled appropriately for the given sample size.
    """

    mu_0: float = field()
    sigma: float = field()
    lmda: float = field()
    L: float = field()

    steady_state: bool = field(default=False, repr=False)
    name: str = field(default='EWMA', repr=False)

    def statistic(self, samples):
        """
        Exponentially weighted average of the samples.

        Samples is first averaged over axis 1, so that when samples.shape[1] is
        2 or more, the statistic is tracking the sample mean.

        Parameters
        ----------
        samples : pandas.DataFrame
            Samples in rows, with one or more columns representing observations
            in each sample.
        """
        samples_z0 = pd.concat([pd.Series(self.mu_0), samples.mean(axis=1)])
        ewm = samples_z0.ewm(alpha=self.lmda, adjust=False).mean()
        return ControlStatistic(
            stat=ewm.iloc[1:],
            nobs=samples.apply(len, axis=1))

    def target(self):
        """
        Desired process mean.
        """
        return self.mu_0

    def lcl(self, nobs):
        """
        Calculates the lower control limits for samples with sizes in `nobs`.

        Parameters
        ----------
        nobs : pandas.Series
            Sample sizes used to construct the statistic at each sample index.

        Notes
        -----
        The argument `nobs` must be the size of the samples used to construct
        the statistic at each index.
        """
        idx = nobs.index.to_series()
        stderr = self.sigma / np.sqrt(nobs)
        if self.steady_state:
            sqrt_term = np.sqrt(self.lmda / (2 - self.lmda))
        else:
            sqrt_term = np.sqrt(self.lmda / (2 - self.lmda) * (1 - (1 - self.lmda)**(2 * idx)))
        return self.mu_0 - self.L * stderr * sqrt_term

    def ucl(self, nobs):
        """
        Calculates the upper control limits for samples with sizes in `nobs`.

        Parameters
        ----------
        nobs : pandas.Series
            Sample sizes used to construct the statistic at each sample index.

        Notes
        -----
        The argument `nobs` must be the size of the samples used to construct
        the statistic at each index.
        """
        idx = nobs.index.to_series()
        stderr = self.sigma / np.sqrt(nobs)
        if self.steady_state:
            sqrt_term = np.sqrt(self.lmda / (2 - self.lmda))
        else:
            sqrt_term = np.sqrt(self.lmda / (2 - self.lmda) * (1 - (1 - self.lmda)**(2 * idx)))
        return self.mu_0 + self.L * stderr * sqrt_term

    @staticmethod
    def from_stddev(mu_0, s_bar, nobs, lmda, L, steady_state=False):
        """
        Constructs EwmaParams from an average sample stddev.

        Parameters
        ----------
        mu_0 : float
            Process target mean.
        s_bar : float
            Average sample stddev from a reference in-control process.
        nobs : int
            Size (fixed) of samples used to calculate (s_bar).
        lmda : float
            Smoothing factor.
        L : float
            Width of the limits in multiples of the stddev of the smoothed mean.
        steady_state : bool, optional
            Whether the process has already decayed to the steady state control limits.
        """
        return EwmaParams(
            mu_0=mu_0,
            sigma=s_bar/c4(nobs),
            lmda=lmda,
            L=L,
            steady_state=steady_state)

    @staticmethod
    def from_data(samples, lmda, L, steady_state=False):
        """
        Constructs an instance of `EwmaParams` from reference samples.

        Reference samples are usually taken from an historical, in-control process.

        Parameters
        ----------
        samples : pandas.DataFrame
            Reference samples where each row is a sample and each column is an
            observation.
        lmda : float
            Smoothing factor.
        L : float
            Width of the limits in multiples of the stddev of the smoothed mean.
        steady_state : bool, optional
            Whether the process has already decayed to the steady state control limits.
        """
        N, nobs = samples.shape
        mu_0 = samples.mean(axis=1).mean()
        s_bar = np.std(samples.values, ddof=1, axis=1).mean()
        return EwmaParams.from_stddev(mu_0, s_bar, nobs, lmda, L, steady_state)

@dataclass
class MewmaParams(ControlParams):
    """
    MEWMA (multivaraite-EMWA) control parameters.

    Tools for calculating the `limit` based on ARL and h4 are provided in
    :func:`mqr.spc.solve_arl` and :func:`mqr.spc.solve_h4`.

    Attributes
    ----------
    mu : array_like
        Desired process output means.
    cov : array_like
        Covariance of process outputs.
    lmda : float
        Smoothing factor.
    limit : float
        Limit for the smoothed statistic, at the steady state.
    name : str, optional
        Name of the control method.

    Notes
    -----
    This class uses dataframe columns to represent process outputs rather than
    observations within a sample. That is, each row contains measurements that
    were taken at the same time (or equivalent), and columns contain measurements
    for the different process dimensions.
    """

    mu: np.ndarray = field(repr=False)
    cov: np.ndarray = field(repr=False)
    lmda: float = field(repr=False)
    limit: float = field(repr=False)

    name: str = field(default='Multivariate EWMA', repr=False)

    def statistic(self, samples):
        """
        Exponentially weighted average of the samples.

        Each row is a sample, and each column is a dimension in the process.

        Parameters
        ----------
        samples : pandas.DataFrame
        """
        init = pd.DataFrame(self.mu[None, :], columns=samples.columns)
        samples_0 = pd.concat([init, samples])
        z = (samples_0 - self.mu).ewm(alpha=self.lmda, adjust=False).mean().iloc[1:, :]
        t2 = pd.Series(index=samples.index)
        for i, idx in enumerate(samples.index):
            t2.iloc[i] = self._t2_stat(z.values[i].T, idx)
        return ControlStatistic(
            stat=t2,
            nobs=samples.apply(len, axis=1))

    def target(self):
        """
        Desired distance (a function of the norm) of the process from `mu`.

        Always zero.
        """
        return 0

    def lcl(self, nobs):
        """
        Calculates the lower control limits for samples with sizes in `nobs`.

        Always None, since MEWMA tracks a norm, which is always non-negative.
        """
        return None

    def ucl(self, nobs):
        """
        Calculates the upper control limits for samples with sizes in `nobs`.
        """
        return pd.Series(self.limit, index=nobs.index)

    def _cov_z(self, i):
        return self.lmda / (2 - self.lmda) * (1 - (1 - self.lmda)**(2 * i)) * self.cov

    def _t2_stat(self, z, i):
        return z @ np.linalg.inv(self._cov_z(i)) @ z[:, None]

    @staticmethod
    def from_data(samples, limit, lmda):
        """
        Constructs MewmaParams from reference samples.

        Reference samples are usually taken from an historical, in-control process.

        Parameters
        ----------
        samples : pandas.DataFrame
            Reference samples, with a sample per row and process dimensions in
            columns.
        limit : float
            Upper control limit. See :func:`mqr.spc.solve_arl` and
            :func:`mqr.spc.solve_h4`, which are tools for calculating this limit.
        lmda : float
            Smoothing factor.
        """
        mu = samples.mean(axis=0).values
        cov = samples.cov(ddof=1).values
        return MewmaParams(mu, cov, lmda, limit)



import mqr.inference.lib.proportion as proportion

from mqr.inference.confint import ConfidenceInterval
from mqr.inference.hyptest import HypothesisTest
from mqr.inference.power import TestPower

import mqr.inference.lib.util as util
import mqr.interop.inference as interop
from mqr.utils import clip_where

import numpy as np
import scipy
import statsmodels

def power_1sample(pa, H0_prop, nobs, alpha=0.05, alternative='two-sided', method='norm-approx'):
    """
    Calculate power of a test of proportion in a sample.

    Null-hypothesis
        `pa` - `H0_prop` == 0

    Parameters
    ----------
    pa : float
        Alternative hypothesis proportion, forming effect size.
    H0_prop : float
        Null-hypothesis proportion.
    nobs : int
        Number of observations.
    alpha : float
        Required significance.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'norm-approx'}, optional
        Test method. Only 'norm-approx', the normal approximation to the
        binomial distribution is implemented.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    """
    if method == 'norm-approx':
        diff = H0_prop - pa
        mu = np.sqrt(H0_prop * (1 - H0_prop) / nobs)
        sigma = np.sqrt(pa * (1 - pa) / nobs)
        dist = scipy.stats.norm()

        if alternative == 'less':
            z = dist.ppf(1 - alpha)
            power = dist.cdf((diff - z * mu) / sigma)
        elif alternative == 'greater':
            z = dist.ppf(1 - alpha)
            power = 1 - dist.cdf((diff + z * mu) / sigma)
        elif alternative == 'two-sided':
            z = dist.ppf(1 - alpha/2)
            power = 1 - dist.cdf((diff + z * mu) / sigma) + dist.cdf((diff - z * mu) / sigma)
        else:
            raise ValueError(util.alternative_error_msg(alternative))
    else:
        raise ValueError(util.method_error_msg(method, ['norm-approx']))

    return TestPower(
        name='proportion',
        alpha=alpha,
        beta=1-power,
        effect=f'{pa:g} - {H0_prop:g} = {pa-H0_prop:g}',
        alternative=alternative,
        method=method,
        sample_size=nobs)

def power_2sample(p1, p2, nobs, alpha=0.05, alternative='two-sided', method='norm-approx'):
    """
    Calculate power of a test of difference of two proportions in two samples.

    Null-hypothesis
        `p1` - `p2` == 0

    Parameters
    ----------
    p1 : float
        First proportion.
    p2 : float
        Second proportion.
    nobs : int
        Number of observations.
    alpha : float
        Required significance.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'norm-approx'}, optional
        Test method. Only 'norm-approx', the normal approximation to the
        binomial distribution, is implemented.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    """
    if method == 'norm-approx':
        diff = p2 - p1
        pavg = (p1 + p2) / 2
        pavg_scale = np.sqrt(2 * pavg * (1 - pavg) / nobs)
        p_scale = np.sqrt(p1 * (1 - p1) / nobs + p2 * (1 - p2) / nobs)
        dist = scipy.stats.norm()

        if alternative == 'less':
            z = dist.ppf(1 - alpha)
            num = diff - z * pavg_scale
            den = p_scale
            power = dist.cdf(num / den)
        elif alternative == 'greater':
            z = dist.ppf(1 - alpha)
            num = diff + z * pavg_scale
            den = p_scale
            power = 1 - dist.cdf(num / den)
        elif alternative == 'two-sided':
            z = dist.ppf(1 - alpha/2)
            num1 = diff + z * pavg_scale
            num2 = diff - z * pavg_scale
            den = p_scale
            power = 1 - dist.cdf(num1 / den) + dist.cdf(num2 / den)
        else:
            raise ValueError(util.alternative_error_msg(alternative))
    else:
        raise ValueError(util.method_error_msg(method, ['norm-approx']))

    return TestPower(
        name='difference between proportions',
        alpha=alpha,
        beta=1-power,
        effect=f'{p1:g} - {p2:g} = {p1-p2:g}',
        alternative=alternative,
        method=method,
        sample_size=nobs)

def size_1sample(pa, H0_prop, alpha, beta, alternative='two-sided', method='norm-approx'):
    """
    Calculate sample size for test of proportion.

    Null-hypothesis
        `pa` - `H0_prop` == 0

    Parameters
    ----------
    pa : float
        Alternative proportion, forming the effect size with `p0`.
    p0 : float
        Null proportion.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'norm-approx', 'arcsin'}, optional
        | 'norm-approx'
        |   Normal approximation to the binomial distribution.
            Solves :func:`power_1sample` equal to requested power by changing `nobs`.
        | 'arcsin'
        |   Inverse-sine approximation to the binomial distribution.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    """
    if method == 'norm-approx':
        def power_fn(nobs):
            return power_1sample(
                pa=pa,
                H0_prop=H0_prop,
                nobs=nobs,
                alpha=alpha,
                alternative=alternative,
                method=method).beta - beta
        nobs_opt = scipy.optimize.fsolve(power_fn, 1)[0]
    elif (method == 'arcsin') or (method == 'invsin-approx'):
        if alternative == 'less' or alternative == 'greater':
            crit = alpha
        elif alternative == 'two-sided':
            crit = alpha / 2
        else:
            raise ValueError(util.alternative_error_msg(alternative))
        dist = scipy.stats.norm()

        Zb = dist.ppf(1-beta)
        Za = -dist.ppf(crit)
        num = Za + Zb
        denom = np.arcsin(np.sqrt(pa)) - np.arcsin(np.sqrt(H0_prop))
        nobs_opt = num**2 / denom**2 / 4
    else:
        raise ValueError(util.method_error_msg(method, ['norm-approx', 'arcsin']))

    return TestPower(
        name='proportion',
        alpha=alpha,
        beta=beta,
        effect=f'{pa:g} - {H0_prop:g} = {pa-H0_prop:g}',
        alternative=alternative,
        method=method,
        sample_size=nobs_opt)

def size_2sample(p1, p2, alpha, beta, alternative='two-sided', method='norm-approx'):
    """
    Calculate sample size to test equality of two proportions.

    Null-hypothesis
        `p1` - `p2` == 0

    Parameters
    ----------
    p1 : float
        First proportion.
    p2 : float
        Second propotion, forming effect size with `p1`.
    alpha : float
        Required significance.
    beta : float
        Required beta (1 - power).
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'norm-approx', 'arcsin'}, optional
        | 'norm-approx'
        |   Numerically solves the normal approximation to the binomial
            distribution by searching over `nobs`.
        | 'arcsin'
        |   Inverse-sine method.

    Returns
    -------
    :class:`mqr.inference.power.TestPower`
    """
    if method == 'norm-approx':
        def power_fn(nobs):
            return power_2sample(
                p1=p1,
                p2=p2,
                nobs=nobs,
                alpha=alpha,
                alternative=alternative,
                method=method).beta - beta
        nobs_opt = scipy.optimize.fsolve(power_fn, 1)[0]
    elif (method == 'arcsin') or (method == 'invsin-approx'):
        if alternative == 'less' or alternative == 'greater':
            crit = alpha
        elif alternative == 'two-sided':
            crit = alpha / 2
        else:
            raise ValueError(util.alternative_error_msg(alternative))
        dist = scipy.stats.norm()

        Zb = dist.ppf(1-beta)
        Za = -dist.ppf(crit)
        num = Za + Zb
        denom = np.arcsin(np.sqrt(p1)) - np.arcsin(np.sqrt(p2))
        nobs_opt = 2 * num**2 / denom**2 / 4
    else:
        raise ValueError(util.method_error_msg(method, ['norm-approx', 'arcsin']))

    return TestPower(
        name='proportion',
        alpha=alpha,
        beta=beta,
        effect=f'{p1:g} - {p2:g} = {p1-p2:g}',
        alternative=alternative,
        method=method,
        sample_size=nobs_opt)

def confint_1sample(count, nobs, conf=0.95, bounded='both', method='agresti-coull'):
    """
    Confidence interval for proportion `count / nobs`.

    Following [1]_, use the 'wilson-cc' or 'jeffreys' method for small sample
    size, and use the 'agresti-coull' or 'jeffreys' for larger sample sizes.
    The authors recommend n = 40 as the boundary between small and large.

    Parameters
    ----------
    count : int
        Number of "true" observations.
    nobs : int
        Total observations.
    conf : float, optional
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}, optional
        Which sides of the interval to close.
    method : {'agresti-coull', 'jeffreys', 'wilson', 'wilson-cc'}, optional
        | 'agresti-coull'
        |   Agresti-Coull interval, see [1]_.
        | 'jeffreys'
        |   Jeffreys interval (a Bayesian method), see [1]_.
        | 'wilson'
        |   Wilson method without continuity correction (method 3 in [2]_).
        | 'wilson-cc'
        |   Wilson method with continuity correction (method 4 in [2]_).
        | (other)
        |   Everything else is passed to
            :func:`sm..proportion_confint <statsmodels.stats.proportion.proportion_confint>`.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Brown, L. D. Cai, T. T. and DasGupta, A. (2001).
            Interval estimation for a binomial proportion.
            Statistical Science, 16(2), 101-133.
    .. [2]  Newcombe, R. G. (1998).
            Two‐sided confidence intervals for the single proportion:
            comparison of seven methods.
            Statistics in medicine, 17(8), 857-872.
    """
    alpha = 1 - conf

    if method == 'agresti-coull':
        lower, upper = proportion.confint_1sample_agresti_coull(count, nobs, conf, bounded)
    elif method == 'jeffreys':
        lower, upper = proportion.confint_1sample_jeffreys(count, nobs, conf, bounded)
    elif method == 'wilson':
        lower, upper = proportion.confint_1sample_wilson(count, nobs, conf, bounded)
    elif method == 'wilson-cc':
        lower, upper = proportion.confint_1sample_wilson_cc(count, nobs, conf, bounded)
    else:
        if bounded == 'both':
            (lower, upper) = statsmodels.stats.proportion.proportion_confint(
                count=count,
                nobs=nobs,
                alpha=alpha,
                method=method)
        else:
            raise ValueError(
                f'Method "{method}" is passed to statsmodels which does not implement '
                'one-sided bounds. ' +
                util.method_error_msg(method, ['agresti-coull', 'jeffreys', 'wilson-cc']))
    value = count / nobs
    return ConfidenceInterval(
        name='proportion',
        method=method,
        value=value,
        lower=lower,
        upper=upper,
        conf=conf,
        bounded=bounded)

def confint_2sample(count1, nobs1, count2, nobs2, conf=0.95, bounded='both', method='newcomb-cc'):
    """
    Confidence interval for difference between proportions `count1 / nobs1 - count2 / nobs2`.

    Parameters
    ----------
    count1 : int
        Number of "true" observations in first sample.
    nobs1 : int
        Total observations in first sample.
    count2 : int
        Number of "true" observations in second sample.
    nobs2 : int
        Total observations in second sample.
    conf : float, optional
        Confidence level that determines the width of the interval.
    bounded : {'both', 'below', 'above'}, optional
        Which sides of the interval to close.
    method : {'agresti-caffo', 'newcomb', 'newcomb-cc'}, optional
        | 'agresti-caffo' (or 'adj-wald')
        |   Agresti-caffo method an adjusted normal approximation.
            See reference [1]_.
        | 'newcomb'
        |   A method presented by Newcomb as method 10 in [2]_.
        | 'newcomb-cc'
        |   Continuity-corrected version of 'newcomb'; method 11 in [2]_.
        | (other)
        |   Everything else is passed to
            :func:`sm..confint_proportions_2indep <statsmodels.stats.proportion.confint_proportions_2indep>`
            for comparison of the difference.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Agresti, A., & Caffo, B. (2000).
            Simple and effective confidence intervals for proportions and differences
            of proportions result from adding two successes and two failures.
            The American Statistician, 54(4), 280-288.
    .. [2]  Newcombe, R. G. (1998).
            Interval estimation for the difference between independent proportions:
            comparison of eleven methods.
            Statistics in medicine, 17(8), 873-890.
    """
    if (method == 'agresti-caffo') or (method == 'adj-wald'):
        lower, upper = proportion.confint_2sample_agresti_caffo(count1, nobs1, count2, nobs2, conf, bounded)
    elif method == 'newcomb':
        lower, upper = proportion.confint_2sample_newcomb(count1, nobs1, count2, nobs2, conf, bounded)
    elif method == 'newcomb-cc':
        lower, upper = proportion.confint_2sample_newcomb_cc(count1, nobs1, count2, nobs2, conf, bounded)
    else:
        if bounded == 'both':
            lower, upper = statsmodels.stats.proportion.confint_proportions_2indep(
                count1, nobs1,
                count2, nobs2,
                alpha=1-conf,
                compare='diff',
                method=method)
        else:
            raise ValueError(
                f'Method "{method}" is passed to statsmodels which does not implement '
                'one-sided bounds. ' +
                util.method_error_msg(method, ['agresti-caffo', 'newcomb-cc']))

    value = count1 / nobs1 - count2 / nobs2
    return ConfidenceInterval(
        name='difference between proportions',
        method=method,
        value=value,
        lower=lower,
        upper=upper,
        conf=conf,
        bounded=bounded)

def test_1sample(count, nobs, H0_prop, alternative='two-sided', method='binom'):
    """
    Hypothesis test for the proportion of "true" elements in a sample.

    Null-hypothesis
        `count` / `nobs` == `H0_prop`

    Parameters
    ----------
    count : int
        Number of "true" observations.
    nobs : int
        Total number of observations.
    H0_prop : float
        Null-hypothesis proportion.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis. Valid for methods 'binom' and 'z';
        method 'chi2' supports only two-sided tests.
    method : {'binom', 'chi2', 'z'}, optional
        | 'binom'
        |   Calls :func:`sm..binom_test <statsmodels.stats.proportion.binom_test>`.
        | 'chi2'
        |   Calls :func:`sm..proportions_chisquare <statsmodels.stats.proportion.proportions_chisquare>`.
        | 'z'
        |   Calls :func:`sm..proportions_ztest <statsmodels.stats.proportion.proportions_ztest>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    alt = interop.alternative(alternative, lib='statsmodels')
    if method == 'binom':
        pvalue = statsmodels.stats.proportion.binom_test(
            count=count,
            nobs=nobs,
            prop=H0_prop,
            alternative=alt)
        stat = np.nan
    elif method == 'z':
        stat, pvalue = statsmodels.stats.proportion.proportions_ztest(
            count=count,
            nobs=nobs,
            value=H0_prop,
            alternative=alt)
    elif method == 'chi2':
        if alternative != 'two-sided':
            raise ValueError(
                f'Method "{method}" is passed to statsmodels which does not implement '
                'one-sided alternatives. ' +
                util.method_error_msg(method, ['binom', 'z']))
        stat, pvalue, _ = statsmodels.stats.proportion.proportions_chisquare(
            count=count,
            nobs=nobs,
            value=H0_prop)
    else:
        raise ValueError(util.method_error_msg(method, ['binom', 'chi2', 'z']))

    return HypothesisTest(
        description='proportion of "true" elements',
        alternative=alternative,
        method=method,
        sample_stat=f'count / nobs',
        sample_stat_target=H0_prop,
        sample_stat_value=count/nobs,
        stat=stat,
        pvalue=pvalue,)

def test_2sample(count1, nobs1, count2, nobs2, H0_diff=0.0, alternative='two-sided', method='agresti-caffo'):
    """
    Hypothesis test for the difference between proportions of two samples.

    Null-hypothesis
        `count1` / `nobs1` - `count2` / `nobs2` == `H0_diff`

    Parameters
    ----------
    count1 : int
        Number of "true" observations in first sample.
    nobs1 : int
        Total number of observations in second sample.
    count2 : int
        Number of "true" observations in second sample.
    nobs2 : int
        Total number of observations in seconds sample.
    H0_diff : float
        Null-hypothesis difference.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : str, optional
        Passed to :func:`sm..test_proportions_2indep <statsmodels.stats.proportion.test_proportions_2indep>`.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`
    """
    alt = interop.alternative(alternative, 'statsmodels')
    res = statsmodels.stats.proportion.test_proportions_2indep(
        count1, nobs1,
        count2, nobs2,
        alternative=alt,
        value=H0_diff,
        method=method)

    return HypothesisTest(
        description='difference between proportions of "true" elements',
        alternative=alternative,
        method=res.method,
        sample_stat=f'count1 / nobs1 - count2 / nobs2',
        sample_stat_target=H0_diff,
        sample_stat_value=count1/nobs1-count2/nobs2,
        stat=res.statistic,
        pvalue=res.pvalue,)



'''
Check call-throughs.
'''

import numbers
import numpy as np
import pytest

import mqr

def test_power_1sample():
    p0 = 0.4
    alpha = 0.05
    nobs = 200

    pa = 0.5
    alternative = 'two-sided'
    method = 'norm-approx'
    res = mqr.inference.proportion.power_1sample(pa, p0, nobs, alpha, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == pytest.approx(0.181922, abs=1e-6)
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_size == nobs

    alternative = 'greater'
    res = mqr.inference.proportion.power_1sample(pa, p0, nobs, alpha, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == pytest.approx(0.111839, abs=1e-6)
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_size == nobs

    pa = 0.3
    alternative = 'less'
    res = mqr.inference.proportion.power_1sample(pa, p0, nobs, alpha, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == pytest.approx(0.0921478, abs=1e-6)
    assert res.effect == '0.3 - 0.4 = -0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_size == nobs

def test_power_2sample():
    p2 = 0.4
    alpha = 0.05
    nobs = 200

    p1 = 0.5
    alternative = 'two-sided'
    method = 'norm-approx'
    res = mqr.inference.proportion.power_2sample(p1, p2, nobs, alpha, alternative, method)
    assert res.name == 'difference between proportions'
    assert res.alpha == 0.05
    assert res.beta == pytest.approx(0.479882, abs=1e-6)
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_size == nobs

    alternative = 'greater'
    res = mqr.inference.proportion.power_2sample(p1, p2, nobs, alpha, alternative, method)
    assert res.name == 'difference between proportions'
    assert res.alpha == 0.05
    assert res.beta == pytest.approx(0.356779, abs=1e-6)
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_size == nobs

    p1 = 0.3
    alternative = 'less'
    res = mqr.inference.proportion.power_2sample(p1, p2, nobs, alpha, alternative, method)
    assert res.name == 'difference between proportions'
    assert res.alpha == 0.05
    assert res.beta == pytest.approx(0.324836, abs=1e-6)
    assert res.effect == '0.3 - 0.4 = -0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert res.sample_size == nobs

def test_size_1sample():
    p0 = 0.4
    alpha = 0.05
    beta = 0.05

    pa = 0.5
    alternative = 'two-sided'
    method = 'norm-approx'
    res = mqr.inference.proportion.size_1sample(pa, p0, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 318

    alternative = 'greater'
    res = mqr.inference.proportion.size_1sample(pa, p0, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 266

    pa = 0.3
    alternative = 'less'
    res = mqr.inference.proportion.size_1sample(pa, p0, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.3 - 0.4 = -0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 244

    pa = 0.5
    alternative = 'two-sided'
    method = 'arcsin'
    res = mqr.inference.proportion.size_1sample(pa, p0, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 321

    alternative = 'greater'
    res = mqr.inference.proportion.size_1sample(pa, p0, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 267

    pa = 0.3
    alternative = 'less'
    res = mqr.inference.proportion.size_1sample(pa, p0, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.3 - 0.4 = -0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 246

def test_size_2sample():
    p2 = 0.4
    alpha = 0.05
    beta = 0.05

    p1 = 0.5
    alternative = 'two-sided'
    method = 'norm-approx'
    res = mqr.inference.proportion.size_2sample(p1, p2, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 641

    alternative = 'greater'
    res = mqr.inference.proportion.size_2sample(p1, p2, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 533

    p1 = 0.3
    alternative = 'less'
    res = mqr.inference.proportion.size_2sample(p1, p2, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.3 - 0.4 = -0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 490

    p1 = 0.5
    alternative = 'two-sided'
    method = 'arcsin'
    res = mqr.inference.proportion.size_2sample(p1, p2, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 642

    alternative = 'greater'
    res = mqr.inference.proportion.size_2sample(p1, p2, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.5 - 0.4 = 0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 534

    p1 = 0.3
    alternative = 'less'
    res = mqr.inference.proportion.size_2sample(p1, p2, alpha, beta, alternative, method)
    assert res.name == 'proportion'
    assert res.alpha == 0.05
    assert res.beta == 0.05
    assert res.effect == '0.3 - 0.4 = -0.1'
    assert res.alternative == alternative
    assert res.method == method
    assert np.ceil(res.sample_size) == 491

def test_confint_1sample():
    count = 5
    nobs = 10
    conf = 0.90
    method = 'beta'
    bounded = 'both'

    res = mqr.inference.proportion.confint_1sample(count, nobs, conf, method=method)
    assert res.name == 'proportion'
    assert res.method == 'beta'
    assert res.value == count / nobs
    assert isinstance(res.lower, numbers.Number)
    assert isinstance(res.upper, numbers.Number)
    assert res.conf == conf
    assert res.bounded == bounded

    count = np.array([0, 20, 40])
    nobs = 40

    bounded = 'both'
    lowers, uppers = list(mqr.inference.proportion.confint_1sample(count, nobs, bounded=bounded, method='agresti-coull'))
    assert lowers == pytest.approx(np.array([0.        , 0.35199527, 0.89560378]))
    assert uppers == pytest.approx(np.array([0.10439622, 0.64800473, 1.        ]))
    lowers, uppers = list(mqr.inference.proportion.confint_1sample(count, nobs, bounded=bounded, method='jeffreys'))
    assert lowers == pytest.approx(np.array([0.        , 0.3496035 , 0.93950202]))
    assert uppers == pytest.approx(np.array([0.06049798, 0.6503965 , 1.        ]))
    lowers, uppers = list(mqr.inference.proportion.confint_1sample(count, nobs, bounded=bounded, method='wilson-cc'))
    assert lowers == pytest.approx(np.array([0.        , 0.34063274, 0.89087533]))
    assert uppers == pytest.approx(np.array([0.10912467, 0.65936726, 1.        ]))

    bounded = 'below'
    lower, upper = mqr.inference.proportion.confint_1sample(count, nobs, bounded=bounded, method='agresti-coull')
    assert lower == pytest.approx(np.array([0.        , 0.37414945, 0.92424085]))
    assert upper == pytest.approx(np.array([1., 1., 1.]))
    lower, upper = mqr.inference.proportion.confint_1sample(count, nobs, bounded=bounded, method='jeffreys')
    assert lower == pytest.approx(np.array([0.        , 0.37290554, 0.95340272]))
    assert upper == pytest.approx(np.array([1., 1., 1.]))
    lower, upper = mqr.inference.proportion.confint_1sample(count, nobs, bounded=bounded, method='wilson-cc')
    assert lower == pytest.approx(np.array([0.        , 0.36247821, 0.91495118]))
    assert upper == pytest.approx(np.array([1., 1., 1.]))

    bounded = 'above'
    lower, upper = mqr.inference.proportion.confint_1sample(count, nobs, bounded=bounded, method='agresti-coull')
    assert lower == pytest.approx(np.array([0., 0., 0.]))
    assert upper == pytest.approx(np.array([0.07575915, 0.62585055, 1.        ]))
    lower, upper = mqr.inference.proportion.confint_1sample(count, nobs, bounded=bounded, method='jeffreys')
    assert lower == pytest.approx(np.array([0., 0., 0.]))
    assert upper == pytest.approx(np.array([0.04659728, 0.62709446, 1.        ]))
    lower, upper = mqr.inference.proportion.confint_1sample(count, nobs, bounded=bounded, method='wilson-cc')
    assert lower == pytest.approx(np.array([0., 0., 0.]))
    assert upper == pytest.approx(np.array([0.08504882, 0.63752179, 1.        ]))

def test_confint_1sample_newcomb():
    """
    Checks values against Table I in [1]_.

    References
    ----------
    .. [1] Newcombe, R. G. (1998).
       Two‐sided confidence intervals for the single proportion:
       comparison of seven methods. Statistics in medicine, 17(8), 857-872.
    """
    from mqr.inference.lib.proportion import confint_1sample_wilson as wilson
    from mqr.inference.lib.proportion import confint_1sample_wilson_cc as wilson_cc

    lower, upper = wilson(81, 263, 0.95, 'both')
    assert lower == pytest.approx(0.2553, abs=1e-4)
    assert upper == pytest.approx(0.3662, abs=1e-4)
    lower, upper = wilson_cc(81, 263, 0.95, 'both')
    assert lower == pytest.approx(0.2535, abs=1e-4)
    assert upper == pytest.approx(0.3682, abs=1e-4)

    lower, upper = wilson(15, 148, 0.95, 'both')
    assert lower == pytest.approx(0.0624, abs=1e-4)
    assert upper == pytest.approx(0.1605, abs=1e-4)
    lower, upper = wilson_cc(15, 148, 0.95, 'both')
    assert lower == pytest.approx(0.0598, abs=1e-4)
    assert upper == pytest.approx(0.1644, abs=1e-4)

    lower, upper = wilson(0, 20, 0.95, 'both')
    assert lower == pytest.approx(0.0000, abs=1e-4)
    assert upper == pytest.approx(0.1611, abs=1e-4)
    lower, upper = wilson_cc(0, 20, 0.95, 'both')
    assert lower == pytest.approx(0.0000, abs=1e-4)
    assert upper == pytest.approx(0.2005, abs=1e-4)

    lower, upper = wilson(1, 29, 0.95, 'both')
    assert lower == pytest.approx(0.0061, abs=1e-4)
    assert upper == pytest.approx(0.1718, abs=1e-4)
    lower, upper = wilson_cc(1, 29, 0.95, 'both')
    assert lower == pytest.approx(0.0018, abs=1e-4)
    assert upper == pytest.approx(0.1963, abs=1e-4)

def test_confint_1sample_brown_cai_dasgupta():
    """
    Checks a few values against Table 5 in [1]_.

    References
    ----------
    [1] Brown, L. D. Cai, T. T. and DasGupta, A. (2001).
        "Interval estimation for a binomial proportion".
        Statistical Science, 16(2), 101-133.
    """
    from mqr.inference.lib.proportion import confint_1sample_jeffreys as jeffreys

    lower, upper = jeffreys(0, 7, 0.95, 'both')
    assert lower == 0.0
    assert upper == pytest.approx(0.292, abs=1e-3)

    lower, upper = jeffreys(3, 10, 0.95, 'both')
    assert lower == pytest.approx(0.093, abs=1e-3)
    assert upper == pytest.approx(0.606, abs=1e-3)

    lower, upper = jeffreys(1, 27, 0.95, 'both')
    assert lower == pytest.approx(0.004, abs=1e-3)
    assert upper == pytest.approx(0.160, abs=1e-3)

    lower, upper = jeffreys(13, 27, 0.95, 'both')
    assert lower == pytest.approx(0.303, abs=1e-3)
    assert upper == pytest.approx(0.664, abs=1e-3)

    lower, upper = jeffreys(15, 30, 0.95, 'both')
    assert lower == pytest.approx(0.328, abs=1e-3)
    assert upper == pytest.approx(0.672, abs=1e-3)

def test_confint_2sample():
    count1 = 5
    nobs1 = 10
    count2 = 15
    nobs2 = 30
    conf = 0.90
    bounded = 'both'

    res = mqr.inference.proportion.confint_2sample(count1, nobs1, count2, nobs2, conf, method='newcomb')
    assert res.name == 'difference between proportions'
    assert res.method == 'newcomb'
    assert res.value == count1 / nobs1 - count2 / nobs2
    assert isinstance(res.lower, numbers.Number)
    assert isinstance(res.upper, numbers.Number)
    assert res.conf == conf
    assert res.bounded == bounded

    conf = 0.95
    count1 = np.array([0, 30, 55, 60])
    nobs1 = 60
    count2 = np.array([0, 20, 30, 40])
    nobs2 = 40

    bounded = 'both'
    lower, upper = mqr.inference.proportion.confint_2sample(count1, nobs1, count2, nobs2, bounded=bounded, method='agresti-caffo')
    assert lower == pytest.approx(np.array([-0.06343951, -0.19584581,  0.0131549 , -0.04807853]))
    assert upper == pytest.approx(np.array([0.04807853, 0.19584581, 0.31710624, 0.06343951]))
    lower, upper = mqr.inference.proportion.confint_2sample(count1, nobs1, count2, nobs2, bounded=bounded, method='newcomb-cc')
    assert lower == pytest.approx(np.array([-0.10912467, -0.20595924,  0.00715208, -0.07496504]))
    assert upper == pytest.approx(np.array([0.07496504, 0.20595924, 0.33993579, 0.10912467]))

    bounded = 'below'
    lower, upper = mqr.inference.proportion.confint_2sample(count1, nobs1, count2, nobs2, bounded=bounded, method='agresti-caffo')
    assert lower == pytest.approx(np.array([-0.05447494, -0.16435898,  0.03758857, -0.03911395]))
    assert upper == pytest.approx(np.array([np.inf, np.inf, np.inf, np.inf]))
    lower, upper = mqr.inference.proportion.confint_2sample(count1, nobs1, count2, nobs2, bounded=bounded, method='newcomb-cc')
    assert lower == pytest.approx(np.array([-0.08504882, -0.17724485,  0.02995461, -0.05794766]))
    assert upper == pytest.approx(np.array([np.inf, np.inf, np.inf, np.inf]))

    bounded = 'above'
    lower, upper = mqr.inference.proportion.confint_2sample(count1, nobs1, count2, nobs2, bounded=bounded, method='agresti-caffo')
    assert lower == pytest.approx(np.array([-np.inf, -np.inf, -np.inf, -np.inf]))
    assert upper == pytest.approx(np.array([0.03911395, 0.16435898, 0.29267257, 0.05447494]))
    lower, upper = mqr.inference.proportion.confint_2sample(count1, nobs1, count2, nobs2, bounded=bounded, method='newcomb-cc')
    assert lower == pytest.approx(np.array([-np.inf, -np.inf, -np.inf, -np.inf]))
    assert upper == pytest.approx(np.array([0.05794766, 0.17724485, 0.31391672, 0.08504882]))

def test_test_1sample():
    count = 5
    nobs = 10
    H0_prop = 0.6
    alternative = 'two-sided'

    res = mqr.inference.proportion.test_1sample(count, nobs, H0_prop, alternative, 'binom')
    assert res.description == 'proportion of "true" elements'
    assert res.alternative == alternative
    assert res.method == 'binom'
    assert res.sample_stat == 'count / nobs'
    assert res.sample_stat_target == H0_prop
    assert res.sample_stat_value == count / nobs
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    res = mqr.inference.proportion.test_1sample(count, nobs, H0_prop, alternative, 'z')
    assert res.description == 'proportion of "true" elements'
    assert res.alternative == alternative
    assert res.method == 'z'
    assert res.sample_stat == 'count / nobs'
    assert res.sample_stat_target == H0_prop
    assert res.sample_stat_value == count / nobs
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

    res = mqr.inference.proportion.test_1sample(count, nobs, H0_prop, alternative, 'chi2')
    assert res.description == 'proportion of "true" elements'
    assert res.alternative == alternative
    assert res.method == 'chi2'
    assert res.sample_stat == 'count / nobs'
    assert res.sample_stat_target == H0_prop
    assert res.sample_stat_value == count / nobs
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

def test_test_2sample():
    count1 = 5
    nobs1 = 10
    count2 = 15
    nobs2 = 60
    H0_diff = 0.5 - 0.25
    alternative = 'two-sided'

    res = mqr.inference.proportion.test_2sample(count1, nobs1, count2, nobs2, H0_diff, alternative)
    assert res.description == 'difference between proportions of "true" elements'
    assert res.alternative == alternative
    assert res.method == 'agresti-caffo'
    assert res.sample_stat == 'count1 / nobs1 - count2 / nobs2'
    assert res.sample_stat_target == H0_diff
    assert res.sample_stat_value == count1 / nobs1 - count2 / nobs2
    assert isinstance(res.stat, numbers.Number)
    assert isinstance(res.pvalue, numbers.Number)

def test_confint_2sample_newcomb():
    """
    Checks values against Table II in [1]_.

    References
    ----------
    .. [1] Newcombe, R. G. (1998).
       Interval estimation for the difference between independent proportions:
       comparison of eleven methods.
       Statistics in medicine, 17(8), 873-890.
    """
    from mqr.inference.lib.proportion import confint_2sample_newcomb as newcomb
    from mqr.inference.lib.proportion import confint_2sample_newcomb_cc as newcomb_cc

    conf = 0.95
    bounded = 'both'

    # (a)
    lower, upper = newcomb(56, 70, 48, 80, conf, bounded)
    assert lower == pytest.approx(0.0524, abs=1e-4)
    assert upper == pytest.approx(0.3339, abs=1e-4)
    lower, upper = newcomb_cc(56, 70, 48, 80, conf, bounded)
    assert lower == pytest.approx(0.0428, abs=1e-4)
    assert upper == pytest.approx(0.3422, abs=1e-4)

    # (b)
    lower, upper = newcomb(9, 10, 3, 10, conf, bounded)
    assert lower == pytest.approx(0.1705, abs=1e-4)
    assert upper == pytest.approx(0.8090, abs=1e-4)
    lower, upper = newcomb_cc(9, 10, 3, 10, conf, bounded)
    assert lower == pytest.approx(0.1013, abs=1e-4)
    assert upper == pytest.approx(0.8387, abs=1e-4)

    # (c)
    lower, upper = newcomb(6, 7, 2, 7, conf, bounded)
    assert lower == pytest.approx(0.0582, abs=1e-4)
    assert upper == pytest.approx(0.8062, abs=1e-4)
    lower, upper = newcomb_cc(6, 7, 2, 7, conf, bounded)
    assert lower == pytest.approx(-0.0290, abs=1e-4)
    assert upper == pytest.approx(0.8423, abs=1e-4)

    # (d)
    lower, upper = newcomb(5, 56, 0, 29, conf, bounded)
    assert lower == pytest.approx(-0.0381, abs=1e-4)
    assert upper == pytest.approx(0.1926, abs=1e-4)
    lower, upper = newcomb_cc(5, 56, 0, 29, conf, bounded)
    assert lower == pytest.approx(-0.0667, abs=1e-4)
    assert upper == pytest.approx(0.2037, abs=1e-4)

    # (e)
    lower, upper = newcomb(0, 10, 0, 20, conf, bounded)
    assert lower == pytest.approx(-0.1611, abs=1e-4)
    assert upper == pytest.approx(0.2775, abs=1e-4)
    lower, upper = newcomb_cc(0, 10, 0, 20, conf, bounded)
    assert lower == pytest.approx(-0.2005, abs=1e-4)
    assert upper == pytest.approx(0.3445, abs=1e-4)

    # (f)
    lower, upper = newcomb(0, 10, 0, 10, conf, bounded)
    assert lower == pytest.approx(-0.2775, abs=1e-4)
    assert upper == pytest.approx(0.2775, abs=1e-4)
    lower, upper = newcomb_cc(0, 10, 0, 10, conf, bounded)
    assert lower == pytest.approx(-0.3445, abs=1e-4)
    assert upper == pytest.approx(0.3445, abs=1e-4)

    # (g)
    lower, upper = newcomb(10, 10, 0, 20, conf, bounded)
    assert lower == pytest.approx(0.6791, abs=1e-4)
    assert upper == pytest.approx(1.0000, abs=1e-4)
    lower, upper = newcomb_cc(10, 10, 0, 20, conf, bounded)
    assert lower == pytest.approx(0.6014, abs=1e-4)
    assert upper == pytest.approx(1.0000, abs=1e-4)

    # (h)
    lower, upper = newcomb(10, 10, 0, 10, conf, bounded)
    assert lower == pytest.approx(0.6075, abs=1e-4)
    assert upper == pytest.approx(1.0000, abs=1e-4)
    lower, upper = newcomb_cc(10, 10, 0, 10, conf, bounded)
    assert lower == pytest.approx(0.5128, abs=1e-4)
    assert upper == pytest.approx(1.0000, abs=1e-4)



from mqr.inference.confint import ConfidenceInterval
from mqr.inference.hyptest import HypothesisTest

import mqr.inference.lib.util as util

import numpy as np
import scipy

def confint(x, y, conf=0.95, bounded='both', method='fisher-z'):
    """
    Confidence interval on the Pearson correlation coefficient.

    Parameters
    ----------
    x, y : array_like
        Calculate correlation between these samples.
    conf : float, optional
        Confidence level that determines the width of the interval.
    bounded : {'both', 'above', 'below'}, optional
        Which sides of the interval to close.
    method : {'fisher-z'}, optional
        Type of statistic. Only Fisher's z-transform (see [1]_), is currently
        implemented.

    Returns
    -------
    :class:`mqr.inference.confint.ConfidenceInterval`

    References
    ----------
    .. [1]  Asuero, A. G., Sayago, A., & González, A. G. (2006).
            The correlation coefficient: An overview.
            Critical reviews in analytical chemistry, 36(1), 41-59.
    """
    if method != 'fisher-z':
        raise ValueError(util.method_error_msg(method, ['fisher-z']))

    if len(x) != len(y):
        raise ValueError(f'Lengths of x and y must be equal.')

    alpha = 1 - conf
    r = scipy.stats.pearsonr(x, y).statistic
    n = len(x)

    corr_mu = _fisher_z(r)
    corr_var = 1 / (n - 3)
    dist = scipy.stats.norm(corr_mu, np.sqrt(corr_var))

    if bounded == 'both':
        lower_z, upper_z = dist.ppf(np.array([alpha/2, 1-alpha/2]))
        lower, upper = _inv_fisher_z(np.array([lower_z, upper_z]))
    elif bounded == 'above':
        lower = -1.0
        upper = _inv_fisher_z(dist.ppf(1-alpha))
    elif bounded == 'below':
        lower = _inv_fisher_z(dist.ppf(alpha))
        upper = 1.0
    else:
        raise ValueError(util.alternative_error_msg(alternative))

    return ConfidenceInterval(
        name='correlation',
        method=method,
        value=r,
        lower=lower,
        upper=upper,
        conf=conf,
        bounded=bounded)

def test(x, y, H0_corr=0.0, alternative='two-sided'):
    """
    Hypothesis test on the Pearson correlation coefficient.

    Null-hypothesis
        corr(`x`, `y`) == `H0_corr`

    Parameters
    ----------
    x, y : array_like
        Test correlation of these two equal-length samples.
    H0_corr : float, optional
        Null-hypothesis correlation coefficient.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Sense of alternative hypothesis.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`

    Raises
    ------
    ValueError
        Length of `x` and `y` are different.

    Notes
    -----
    Chooses a method depending on the null-hypothesis. When the null-hypothesis
    correlation is zero, calculates a t-statistic according to Pearson. On the
    other hand, when the null-hypothesis correlation is non-zero, calculates a
    Gaussian statistic based on Fisher's z-transform (see [1]_).

    References
    ----------
    .. [1]  Asuero, A. G., Sayago, A., & González, A. G. (2006).
            The correlation coefficient: An overview.
            Critical reviews in analytical chemistry, 36(1), 41-59.
    """
    if len(x) != len(y):
        raise ValueError(f'Lengths of x and y must be equal.')

    r = scipy.stats.pearsonr(x, y).statistic
    n = len(x)
    if np.isclose(H0_corr, 0.0):
        method = 'pearson'
        dist = scipy.stats.t(n-2)
        stat = r * np.sqrt((n - 2) / (1 - np.square(r)))
    else:
        method = 'fisher-z'
        mu = _fisher_z(H0_corr)
        var = 1 / (n - 3)
        dist = scipy.stats.norm(mu, np.sqrt(var))
        stat = _fisher_z(r)

    if alternative == 'two-sided':
        pvalue = 2 * np.minimum(dist.cdf(stat), 1-dist.cdf(stat))
    elif alternative == 'less':
        pvalue = dist.cdf(stat)
    elif alternative == 'greater':
        pvalue = 1 - dist.cdf(stat)
    else:
        raise ValueError(util.alternative_error_msg(alternative))

    x_name = util.var_name(x, 'x')
    y_name = util.var_name(y, 'y')
    return HypothesisTest(
        description='correlation coefficient',
        alternative=alternative,
        method=method,
        sample_stat=f'corr({x_name}, {y_name})',
        sample_stat_target=H0_corr,
        sample_stat_value=r,
        stat=stat,
        pvalue=pvalue,)

def test_diff(x1, y1, x2, y2, H0_corr1=0.0, H0_corr2=0.0, alternative='two-sided', method='fisher-z'):
    """
    Hypothesis test on the difference of two Pearson correlations.

    Null-hypothesis
        corr(`x1`, `y1`) - corr(`x2`, `y2`) == `H0_corr1` - `H0_corr2`

    Parameters
    ----------
    x1, y1 : array_like
        First pair of samples.
    x2, y2 : array_like
        Second pair of samples whose correlation will be subtracted from the
        correlation of the first pair to form the difference.
    H0_corr1, H0_corr2 : float, optional
        Null-hypothesis difference is `H0_corr1` - `H0_corr2`.
    alternative : {'both', 'less', 'greater'}, optional
        Sense of alternative hypothesis.
    method : {'fisher-z'}, optional
        Type of statistic; only 'fisher-z' is implemented.

    Returns
    -------
    :class:`mqr.inference.hyptest.HypothesisTest`

    Notes
    -----
    Uses Fisher's z-transform, see [1]_.

    Note that the alternatives mean "less" or "greater" on the real number line,
    not in absolute value.

    Raises
    ------
    ValueError
        Lengths of `x1` and `y1` are different, or lengths of `x2` and `y2` are
        different.

    References
    ----------
    .. [1]  Asuero, A. G., Sayago, A., & González, A. G. (2006).
            The correlation coefficient: An overview.
            Critical reviews in analytical chemistry, 36(1), 41-59.
    """
    if method != 'fisher-z':
        raise ValueError(util.method_error_msg(method, ['fisher-z']))

    if len(x1) != len(y1):
        raise ValueError('Lengths of x1 and y1 must be equal.')
    if len(x2) != len(y2):
        raise ValueError('Lengths of x2 and y2 must be equal.')

    n1 = len(x1)
    n2 = len(x2)
    mu1 = _fisher_z(H0_corr1)
    mu2 = _fisher_z(H0_corr2)
    var1 = 1 / (n1 - 3)
    var2 = 1 / (n2 - 3)
    mu = mu1 - mu2
    var = var1 + var2
    dist = scipy.stats.norm(mu, np.sqrt(var))

    r1 = np.corrcoef(x1, y1)[0, 1]
    r2 = np.corrcoef(x2, y2)[0, 1]
    Z1 = _fisher_z(r1)
    Z2 = _fisher_z(r2)
    Z = Z1 - Z2
    stat = Z

    if alternative == 'two-sided':
        pvalue = 2 * np.minimum(dist.cdf(stat), 1-dist.cdf(stat))
    elif alternative == 'less':
        pvalue = dist.cdf(stat)
    elif alternative == 'greater':
        pvalue = 1 - dist.cdf(stat)
    else:
        raise ValueError(util.alternative_error_msg(alternative))

    x1_name = util.var_name(x1, 'x1')
    y1_name = util.var_name(y1, 'y1')
    x2_name = util.var_name(x2, 'x2')
    y2_name = util.var_name(y2, 'y2')
    return HypothesisTest(
        description='difference between correlation coefficients',
        alternative=alternative,
        method=method,
        sample_stat=f'corr({x1_name}, {y1_name}) - corr({x2_name}, {y2_name})',
        sample_stat_target=H0_corr1-H0_corr2,
        sample_stat_value=r1-r2,
        stat=stat,
        pvalue=pvalue,)

def _fisher_z(r):
    return np.arctanh(r)

def _inv_fisher_z(z):
    return np.tanh(z)



"""
=======================================
Data transforms (:mod:`mqr.transforms`)
=======================================

User guide
    :doc:`/user_guide/data-analysis`

Detailed examples
    https://github.com/nklsxn/mqrpy-guide

.. rubric:: Functions

.. autosummary::
    :toctree: generated/

    zscore
"""
from collections.abc import Iterable
import numpy as np
import pandas as pd

def zscore(data, mean=None, stddev=None):
    '''
    Create functions that z-score the given data, per-column.

    The returned functions (`z` or `z_inv`) can take as input a DataFrame whose
    columns are a subset of the columns in `data`, or a Series whose name is one
    of the columns of `data`.

    Parameters
    ----------
    data : pandas.DataFrame
        Data to z-score. Mean and stddev are calculated for each column.
    mean : float or array_like, optional
        Mean to subtract from the data. Defaults to the mean of `data`.
    stddev : float or array_like, optional
        Stddev to divide out of the data. Defaults to the population sttdev of `data`.

    Returns
    -------
    stats : pandas.DataFrame
        Mean and standard deviation values for each column in data (where
        columns are the same as in data).
    z : callable
        A transform from data to its z-scored values.
    z_inv : callable
        A transform from z-scored values to the original space.
    '''
    stats = pd.DataFrame(index=['mean', 'std'], columns=data.columns)

    if mean is None:
        stats.loc['mean'] = data.apply(np.mean)
    else:
        stats.loc['mean'] = mean

    if stddev is None:
        stats.loc['std'] = data.apply(np.std, ddof=0)
    else:
        stats.loc['std'] = stddev

    def z(col):
        mean, std = stats[col.name]
        return (col - mean) / std
    def z_inv(col):
        """
        """
        mean, std = stats[col.name]
        return col * std + mean
    def _z(x):
        if isinstance(x, pd.DataFrame):
            return x.apply(z)
        elif isinstance(x, pd.Series):
            return z(x)
        else:
            raise ValueError('Pass either a DataFrame or a Series.')
    def _z_inv(x):
        if isinstance(x, pd.DataFrame):
            return x.apply(z_inv)
        elif isinstance(x, pd.Series):
            return z_inv(x)
        else:
            raise ValueError('Pass either a DataFrame of a Series.')
    return stats, _z, _z_inv
